import json
import numpy as np
import pandas as pd
import dataknobs.structures.document as dk_doc
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Set


# Key annotations column name constants for use across annotation interfaces
KEY_START_POS_COL = 'start_pos'
KEY_END_POS_COL = 'end_pos'
KEY_TEXT_COL = 'text'
KEY_ANN_TYPE_COL = 'ann_type'


class AnnotationsMetaData(dk_doc.MetaData):
    '''
    Container for annotations meta-data, identifying key column names.

    NOTE: this object contains only information about annotation column names
    and not annotation table values.
    '''
    def __init__(
            self,
            start_pos_col: str = KEY_START_POS_COL,
            end_pos_col: str = KEY_END_POS_COL,
            text_col: str = KEY_TEXT_COL,
            ann_type_col: str = KEY_ANN_TYPE_COL,
            sort_fields: List[str] = (KEY_START_POS_COL, KEY_END_POS_COL),
            sort_fields_ascending: List[bool] = (True, False),
            **kwargs
    ):
        '''
        Initialize with key (and more) column names and info.

        Key column types:
          * start_pos
          * end_pos
          * text
          * ann_type

        NOTEs:
          * Actual table columns can be named arbitrarily
             * BUT: interactions through annotations classes and interfaces
               relating to the "key" columns must use the key column constants

        :param start_pos_col: Col name for the token starting position
        :param end_pos_col: Col name for the token ending position
        :param text_col: Col name for the token text
        :param ann_type_col: Col name for the annotation types
        :param sort_fields: The col types relevant for sorting annotation rows
        :param sort_fields_ascending: To specify sort order of sort_fields
        :param **kwargs: More column types mapped to column names
        '''
        super().__init__(
            {
                KEY_START_POS_COL: start_pos_col,
                KEY_END_POS_COL: end_pos_col,
                KEY_TEXT_COL: text_col,
                KEY_ANN_TYPE_COL: ann_type_col,
            },
            **kwargs
        )
        self.sort_fields = list(sort_fields)
        self.ascending = sort_fields_ascending

    @property
    def start_pos_col(self) -> str:
        ''' Get the column name for the token starting postition '''
        return self.data[KEY_START_POS_COL]

    @property
    def end_pos_col(self) -> str:
        ''' Get the column name for the token ending position '''
        return self.data[KEY_END_POS_COL]

    @property
    def text_col(self) -> str:
        ''' Get the column name for the token text '''
        return self.data[KEY_TEXT_COL]

    @property
    def ann_type_col(self) -> str:
        ''' Get the column name for the token annotation type '''
        return self.data[KEY_ANN_TYPE_COL]

    def get_col(self, col_type: str, missing: str = None) -> str:
        '''
        Get the name of the column having the given type (including key column
        types but not derived,) or get the missing value.

        :param col_type: The type of column name to get
        :param missing: The value to return for unknown column types
        :return: The column name or the missing value
        '''
        return self.get_value(col_type, missing)

    def sort_df(self, an_df: pd.DataFrame):
        '''
        Sort an annotations dataframe according to this metadata.
        :param an_df: An annotations dataframe
        :return: The sorted annotations dataframe.
        '''
        if self.sort_fields is not None:
            an_df = an_df.sort_values(
                self.sort_fields, ascending=self.ascending
            )
        return an_df


class DerivedAnnotationColumns(ABC):
    '''
    Interface for injecting derived columns into AnnotationsMetaData.
    '''
    @abstractmethod
    def get_col_value(
            self,
            metadata: AnnotationsMetaData,
            col_type: str,
            row: pd.Series,
            missing: str = None,
    ) -> str:
        '''
        Get the value of the column in the given row derived from col_type.

        :param metadata: The AnnotationsMetaData
        :param col_type: The type of column value to derive
        :param row: A row from which to get the value.
        :param missing: The value to return for unknown or missing column
        :return: The row value or the missing value
        '''
        raise NotImplementedError


class AnnotationsRowAccessor:
    '''
    A class that accesses row data according to the metadata and derived cols.
    '''
    def __init__(
            self,
            metadata: AnnotationsMetaData,
            derived_cols: DerivedAnnotationColumns = None
    ):
        '''
        :param metadata: The metadata for annotation columns
        :param derived_cols: A DerivedAnnotationColumns instance for injecting
            derived columns.
        '''
        self.metadata = metadata
        self.derived_cols = derived_cols

    def get_col_value(
            self,
            col_type: str,
            row: pd.Series,
            missing: str = None,
    ) -> str:
        '''
        Get the value of the column in the given row with the given type.

        This gets the value from the first existing column in the row from:
          * The metadata.get_col(col_type) column
          * col_type itself
          * The columns derived from col_type

        :param col_type: The type of column value to get
        :param row: A row from which to get the value.
        :param missing: The value to return for unknown or missing column
        :return: The row value or the missing value
        '''
        value = missing
        col = self.metadata.get_col(col_type, None)
        if col is None or col not in row.index:
            if col_type in self.metadata.data:
                value = row[col_type]
            else:
                if self.derived_cols is not None:
                    value = self.derived_cols.get_col_value(
                        self.metadata, col_type, row, missing
                    )
        else:
            value = row[col]
        return value


class Annotations:
    '''
    DAO for collecting and managing a table of annotations, where each row
    carries annotation information for an input token.

    The data in this class is maintained either as a list of dicts, each dict
    representing a "row," or as a pandas DataFrame, depending on the latest
    access. Changes in either the lists or dataframe will be reflected in the
    alternate data structure.
    '''
    def __init__(
            self,
            metadata: AnnotationsMetaData,
            df: pd.DataFrame = None,
    ):
        '''
        Construct as empty or initialize with the dataframe form.
        :param df: A dataframe with annotation records.
        '''
        self.metadata = metadata
        self._annotations_list = None
        self._df = df

    @property
    def ann_row_dicts(self) -> List[Dict[str, Any]]:
        '''
        Get the annotations as a list of dictionaries.
        '''
        if self._annotations_list is None:
            self._annotations_list = self._build_list()
        return self._annotations_list

    @property
    def df(self) -> pd.DataFrame:
        '''
        Get the annotations as a pandas dataframe.
        '''
        if self._df is None:
            self._df = self._build_df()
        return self._df

    def clear(self) -> pd.DataFrame:
        ''' Clear/empty out all annotations, returning the annotations df '''
        rv = self.df
        self._df = None
        self._annotations_list = None
        return rv

    def is_empty(self) -> bool:
        return (
            (self._df is None or len(self._df) == 0) and
            (self._annotations_list is None or len(self._annotations_list) == 0)
        )

    def add_dict(self, annotation: Dict[str, Any]):
        '''
        Add the annotation dict.
        '''
        self.ann_row_dicts.append(annotation)

    def add_dicts(self, annotations: List[Dict[str, Any]]):
        '''
        Add the annotation dicts.
        '''
        self.ann_row_dicts.extend(annotations)

    def add_df(self, an_df: pd.DataFrame):
        '''
        Add (concatentate) the annotation dataframe to the current annotations.
        '''
        df = self.metadata.sort_df(
            pd.concat([self.df, an_df])
        )
        self.set_df(df)

    def _build_list(self) -> List[Dict[str, Any]]:
        '''
        Build the annotations list from the dataframe.
        '''
        alist = None
        if self._df is not None:
            alist = self._df.to_dict(orient='records')
            self._df = None
        return alist if alist is not None else list()

    def _build_df(self) -> pd.DataFrame:
        '''
        Get the annotations as a df.
        '''
        df = None
        if self._annotations_list is not None:
            if len(self._annotations_list) > 0:
                df = self.metadata.sort_df(
                    pd.DataFrame(self._annotations_list)
                )
            self._annotations_list = None
        return df

    def set_df(self, df: pd.DataFrame):
        '''
        Set (or reset) this annotation's dataframe.
        :param df: The new annotations dataframe.
        '''
        self._df = df
        self._annotations_list = None


class AnnotationsBuilder:
    '''
    A class for building annotations.
    '''

    def __init__(
            self,
            metadata: AnnotationsMetaData,
            data_defaults: Dict[str, Any],
    ):
        '''
        :param metadata: The annotations metadata
        :param data_defaults: Dict[ann_colname, default_value] with default
            values for annotation columns
        '''
        self.metadata = metadata if metadata is not None else AnnotationsMetaData()
        self.data_defaults = data_defaults

    def build_annotation_row(
            self,
            start_pos: int,
            end_pos: int,
            text: str,
            ann_type: str,
            **kwargs
    ) -> Dict[str, Any]:
        '''
        Build an annotation row with the mandatory key values and those from
        the remaining keyword arguments.

        For those kwargs whose names match metadata column names, override the
        data_defaults and add remaining data_default attributes.

        :param result_row_dict: The result row dictionary being built
        :param start_pos: The token start position
        :param end_pos: The token end position
        :param text: The token text
        :param ann_type: The annotation type
        :return: The result_row_dict
        '''
        return self.do_build_row({
            self.metadata.start_pos_col: start_pos,
            self.metadata.end_pos_col: end_pos,
            self.metadata.text_col: text,
            self.metadata.ann_type_col: ann_type,
        }, **kwargs)

    def do_build_row(
            self,
            key_fields: Dict[str, Any],
            **kwargs
    ) -> Dict[str, Any]:
        '''
        Do the row building with the key fields, followed by data defaults,
        followed by any extra kwargs.
        :param key_fields: The dictionary of key fields
        :param kwargs: Any extra fields to add
        '''
        result = dict()
        result.update(key_fields)
        if self.data_defaults is not None:
            # Add data_defaults
            result.update(self.data_defaults)
        if kwargs is not None:
            # Override with extra kwargs
            result.update(kwargs)
        return result


class RowData:
    '''
    A wrapper for an annotation row (pd.Series) to facilitate e.g., grouping.
    '''

    def __init__(
            self,
            metadata: AnnotationsMetaData,
            row: pd.Series,
    ):
        self.metadata = metadata
        self.row = row

    @property
    def loc(self):
        return self.row.name

    def __repr__(self) -> str:
        return f'[{self.start_pos}:{self.end_pos})"{self.text}"'

    @property
    def start_pos(self) -> int:
        return self.row[self.metadata.start_pos_col]

    @property
    def end_pos(self) -> int:
        return self.row[self.metadata.end_pos_col]

    @property
    def text(self) -> str:
        return self.row[self.metadata.text_col]
        
    def is_subset(self, other_row: 'RowData') -> bool:
        '''
        Determine whether this row's span is a subset of the other.
        :param other_row: The other row
        '''
        return (
            self.start_pos >= other_row.start_pos and
            self.end_pos <= other_row.end_pos
        )

    def is_subset_of_any(self, other_rows: List['RowData']) -> bool:
        '''
        Determine whether this row is a subset of any of the others
        according to text span coverage.
        :param other_rows: The rows to test for this to be a subset of any
        '''
        result = False
        for other_row in other_rows:
            if self.is_subset(other_row):
                result = True
                break
        return result


class AnnotationsGroup:
    '''
    Container for annotation rows that belong together as a (consistent) group.

    NOTE: An instance will only accept rows on condition of consistency per its
    acceptance function.
    '''
    def __init__(
            self,
            row_accessor: AnnotationsRowAccessor,
            field_col_type: str,
            accept_fn: Callable[['AnnotationsGroup', RowData], bool],
            group_type: str = None,
            group_num: int = None,
            valid: bool = True,
            autolock: bool = False,
    ):
        '''
        :param row_accessor: The annotations row_accessor
        :param field_col_type: The col_type for the group field_type for retrieval
           using the annotations row accessor
        :param accept_fn: A fn(g, row_data) that returns True to accept the row
            data into this group g, or False to reject the row. If None, then
            all rows are always accepted.
        :param group_type: An optional (override) type for identifying this group.
        :param group_num: An optional number for identifying this group.
        :param valid: True if the group is valid, or False if not
        :param autolock: True to automatically lock this group when (1) at
            least one row has been added and (2) a row is rejected.
        '''
        self.rows = list()   # List[RowData]
        self.row_accessor = row_accessor
        self.field_col_type = field_col_type
        self.accept_fn = accept_fn
        self._group_type = group_type
        self._group_num = group_num
        self._valid = valid
        self._autolock = autolock
        self._locked = False
        self._locs = None    # track loc's for recognizing dupes
        self._key = None     # a hash key using the _locs
        self._df = None
        self._ann_type = None

    @property
    def is_locked(self) -> bool:
        '''
        Get whether this group is locked from adding more rows.
        '''
        return self._locked

    @is_locked.setter
    def is_locked(self, value: bool):
        '''
        Set this group as locked (value=True) or unlocked (value=False) to
        allow or disallow more rows from being added regardless of the accept
        function.

        Note that while unlocked only rows that pass the accept function will
        be added.

        :param value: True to lock or False to unlock this group.
        '''
        self._locked = value

    @property
    def is_valid(self) -> bool:
        '''
        Get whether this group is currently marked as valid.
        '''
        return self._valid

    @is_valid.setter
    def is_valid(self, value: bool):
        '''
        Mark this group as valid (value=True) or invalid (value=False).
        :param value: True for valid or False for invalid.
        '''
        self._valid = value

    @property
    def autolock(self) -> bool:
        '''
        Get whether this group is currently set to autolock.
        '''
        return self._autolock

    @autolock.setter
    def autolock(self, value: bool):
        '''
        Set this group to autolock (True) or not (False).
        :param value: True for False to autolock or not.
        '''
        self._autolock = value

    def __repr__(self):
        return json.dumps(self.to_dict())

    @property
    def size(self) -> int:
        '''
        Get the number of rows in this group.
        '''
        return len(self.rows)

    @property
    def group_type(self) -> str:
        '''
        Get this group's type, which is either an "override" value that has
        been set, or the "ann_type" value of the first row added.
        '''
        return (
            self._group_type
            if self._group_type is not None
            else self.ann_type
        )

    @group_type.setter
    def group_type(self, value: str):
        ''' Set this group's type '''
        self._group_type = value

    @property
    def group_num(self) -> int:
        ''' Get this group's number '''
        return self._group_num

    @group_num.setter
    def group_num(self, value: int):
        ''' Set this group's num '''
        self._group_num = value

    @property
    def df(self) -> pd.DataFrame:
        ''' Get this group as a dataframe '''
        if self._df is None:
            self._df = pd.DataFrame([r.row for r in self.rows])
        return self._df

    @property
    def ann_type(self) -> str:
        ''' Get this record's annotation type '''
        return self._ann_type

    @property
    def text(self) -> str:
        return ' '.join([
            row.text for row in self.rows
        ])

    @property
    def locs(self) -> List[int]:
        if self._locs is None:
            self._locs = [r.loc for r in self.rows]
        return self._locs

    @property
    def key(self) -> str:
        '''
        A hash key for this group.
        '''
        if self._key is None:
            self._key = '_'.join([str(x) for x in sorted(self.locs)])
        return self._key

    def copy(self) -> 'AnnotationsGroup':
        result = AnnotationsGroup(
            self.row_accessor, self.field_col_type, self.accept_fn,
            group_type=self.group_type,
            group_num=self.group_num,
            valid=self.is_valid,
            autolock=self.autolock
        )
        result.rows = self.rows.copy()
        result._locked = self._locked  # pylint: disable=protected-access
        result._ann_type = self._ann_type  # pylint: disable=protected-access

    def add(self, rowdata: RowData) -> bool:
        '''
        Add the row if the group is not locked and the row belongs in this
        group, or return False.

        If autolock is True and a row fails to be added (after the first
        row has been added,) "lock" the group and refuse to accept any more
        rows.
        
        :param rowdata: The row to add
        :return: True if the row belongs and was added; otherwise, False
        '''
        result = False
        if self._locked:
            return result

        if self.accept_fn is None or self.accept_fn(self, rowdata):
            self.rows.append(rowdata)
            self._df = None
            self._locs = None
            self._key = None
            if self._ann_type is None:
                self._ann_type = self.row_accessor.get_col_value(
                    KEY_ANN_TYPE_COL, rowdata.row, missing=None,
                )
            result = True

        if not result and self.size > 0 and self.autolock:
            self._locked = True

        return result

    def to_dict(self) -> Dict[str, str]:
        '''
        Get this group (record) as a dictionary of field type to text values.
        '''
        return {
            self.row_accessor.get_col_value(self.field_col_type): row.text
            for row in self.rows
        }

    def is_subset(self, other: 'AnnotationsGroup') -> bool:
        '''
        Determine whether the this group's text is contained within the others.
        :param other: The other group
        '''
        result = True
        for my_row in self.rows:
            if not my_row.is_subset_of_any(other.rows):
                result = False
                break
        return result

    def is_subset_of_any(
            self,
            groups: List['AnnotationsGroup']
    ) -> 'AnnotationsGroup':
        '''
        Determine whether this group is a subset of any of the given groups.
        :param groups: List of annotation groups
        :return: The first AnnotationsGroup that this group is a subset of, or
            None
        '''
        result = None
        for other_group in groups:
            if self.is_subset(other_group):
                result = other_group
                break
        return result

    def remove_row(
            self,
            row_idx: int,
    ) -> RowData:
        '''
        Remove the row from this group and optionally update the annotations
        accordingly.

        :param row_idx: The positional index of the row to remove
        :return: The removed row data instance
        '''
        rowdata = self.rows.pop(row_idx)

        # Reset cached values
        self._df = None
        self._locs = None
        self._key = None
        
        return rowdata


class MergeStrategy(ABC):
    '''
    A merge strategy to be injected based on entity types being merged.
    '''
    @abstractmethod
    def merge(
            self, group: AnnotationsGroup
    ) -> List[Dict[str, Any]]:
        '''
        Process the annotations in the given annotations group, returning the
        group's merged annotation dictionaries.
        '''
        raise NotImplementedError


class PositionalAnnotationsGroup(AnnotationsGroup):
    '''
    Container for annotations that either overlap with each other or don't.
    '''

    def __init__(self, overlap: bool, rectype: str = None, gnum: int = -1):
        '''
        :param overlap: If False, then only accept rows that don't overlap; else
            only accept rows that do ovelap
        '''
        super().__init__(None, None, None, group_type=rectype, group_num=gnum)
        self.overlap = overlap
        self.start_pos = -1
        self.end_pos = -1

    def __repr__(self) -> str:
        return f'nrows={len(self.rows)}[{self.start_pos},{self.end_pos})"{self.entity_text}"'

    @property
    def entity_text(self) -> str:
        jstr = ' | ' if self.overlap else ' '
        return jstr.join(r.entity_text for r in self.rows)

    def belongs(self, rowdata: RowData) -> bool:
        '''
        Determine if the row belongs in this instance based on its overlap
        or not.
        :param rowdata: The rowdata to test
        :return: True if the rowdata belongs in this instance
        '''
        result = True  # Anything belongs to an empty group
        if len(self.rows) > 0:
            start_overlaps = self._is_in_bounds(rowdata.start_pos)
            end_overlaps = self._is_in_bounds(rowdata.end_pos - 1)
            result = start_overlaps or end_overlaps
            if not self.overlap:
                result = not result
        if result:
            if self.start_pos < 0:
                self.start_pos = rowdata.start_pos
                self.end_pos = rowdata.end_pos
            else:
                self.start_pos = min(self.start_pos, rowdata.start_pos)
                self.end_pos = max(self.end_pos, rowdata.end_pos)
        return result

    def _is_in_bounds(self, char_pos):
        return char_pos >= self.start_pos and char_pos < self.end_pos

    def copy(self) -> 'PositionalAnnotationsGroup':
        result = PositionalAnnotationsGroup(self.overlap)
        result.start_pos = self.start_pos
        result.end_pos = self.end_pos
        result.rows = self.rows.copy()
        return result

    #TODO: Add comparison and merge functions


class OverlapGroupIterator:
    '''
    Given:
      * annotation rows (dataframe)
        * in order sorted by
          * start_pos (increasing for input order), and
          * end_pos (decreasing for longest spans first)
    Collect:
      * overlapping consecutive annotations
      * for processing
    '''

    def __init__(self, an_df: pd.DataFrame):
        '''
        :param an_df: An annotations.as_df DataFrame, sliced and sorted.
        '''
        self.an_df = an_df
        self._cur_iter = None
        self._queued_row_data = None
        self.cur_group = None
        self.reset()

    def next_group(self) -> AnnotationsGroup:
        group = None
        if self.has_next:
            group = PositionalAnnotationsGroup(True)
            while self.has_next and group.belongs(self._queued_row_data):
                self._queue_next()
            self.cur_group = group
        return group

    def reset(self):
        self._cur_iter = self.an_df.iterrows()
        self._queue_next()
        self.cur_group = None

    @property
    def has_next(self) -> bool:
        return self._queued_row_data is not None

    def _queue_next(self):
        try:
            _loc, row = next(self._cur_iter)
            self._queued_row_data = RowData(None, row)  #TODO: add metadata
        except StopIteration:
            self._queued_row_data = None


def merge(
        annotations: Annotations,
        merge_strategy: MergeStrategy,
) -> Annotations:
    '''
    Merge the overlapping groups according to the given strategy.
    '''
    og_iter = OverlapGroupIterator(annotations.as_df)
    result = Annotations(annotations.metadata)
    while og_iter.has_next:
        og = og_iter.next_group()
        result.add_dicts(merge_strategy.merge(og))
    return result
        

class AnnotationsGroupList:
    '''
    Container for a list of annotation groups.
    '''
    def __init__(
            self,
            groups: List[AnnotationsGroup] = None,
            accept_fn: Callable[
                ['AnnotationsGroupList', AnnotationsGroup], bool
            ] = lambda l, g: l.size == 0 or not g.is_subset_of_any(l.groups)
    ):
        '''
        :param groups: The initial groups for this list
        :param accept_fn: A fn(l, g) that returns True to accept the group, g,
            into this list, l, or False to reject the group. If None, then all
            groups are always accepted. The default function will reject any
            group that is a subset of any existing group in the list.
        '''
        self.groups = groups if groups is not None else list()
        self.accept_fn = accept_fn
        self._coverage = None

    def __repr__(self) -> str:
        return str(self.groups)

    @property
    def size(self) -> int:
        ''' Get the number of groups in this list '''
        return len(self.groups)

    @property
    def coverage(self) -> int:
        '''
        Get the total number of (token) rows covered by the groups
        '''
        if self._coverage is None:
            locs = set()
            for group in self.groups:
                locs.update(set(group.locs))
            self._coverage = len(locs)
        return self._coverage

    @property
    def df(self) -> pd.DataFrame:
        return pd.DataFrame([r.row for g in self.groups for r in g.rows])

    def copy(self) -> 'AnnotationsGroupList':
        result = AnnotationsGroupList(
            self.groups.copy(), accept_fn=self.accept_fn
        )
        result._coverage = self._coverage  # pylint: disable=protected-access
        return result

    def add(self, group: AnnotationsGroup) -> bool:
        '''
        Add the group if it belongs in this group list or return False.
        :param group: The group to add
        :return: True if the group belongs and was added; otherwise, False
        '''
        result = False
        if self.accept_fn is None or self.accept_fn(self, group):
            self.groups.append(group)
            self._coverage = None
            result = True
        return result

    def is_subset(self, other: 'AnnotationsGroupList') -> bool:
        '''
        Determine whether the this group's text spans are contained within all
        of the other's.
        :param other: The other group list
        '''
        result = True
        for my_group in self.groups:
            if not my_group.is_subset_of_any(other.groups):
                result = False
                break
        return result


class AnnotatedText(dk_doc.Text):
    '''
    A Text object that manages its own annotations.
    '''

    def __init__(
            self,
            text_str: str,
            metadata: dk_doc.TextMetaData = None,
            annots: Annotations = None,
            bookmarks: Dict[str, pd.DataFrame] = None,
            text_obj: dk_doc.Text = None,
            annots_metadata: AnnotationsMetaData = None,
    ):
        '''
        :param text_str: The text string
        :param metadata: The text's metadata
        :param annots: The annotations
        :param bookmarks: The annotation bookmarks
        :param text_obj: A text_obj to override text_str and metadata
            initialization
        :param annots_metadata: Override for default annotations metadata
            (NOTE: ineffectual if an annots instance is provided.)
        '''
        super().__init__(
            text_obj.text if text_obj is not None else text_str,
            text_obj.metadata if text_obj is not None else metadata
        )
        self._annots = annots
        self._bookmarks = bookmarks
        self._annots_metadata = annots_metadata

    @property
    def annotations(self) -> Annotations:
        '''
        Get the this object's annotations
        '''
        if self._annots is None:
            self._annots = Annotations(
                self._annots_metadata or AnnotationsMetaData()
            )
        return self._annots

    @property
    def bookmarks(self) -> Dict[str, pd.DataFrame]:
        '''
        Get this object's bookmarks
        '''
        if self._bookmarks is None:
            self._bookmarks = dict()
        return self._bookmarks

    def get_text(
            self,
            annot2mask: Dict[str, str] = None,
            annot_df: pd.DataFrame = None,
            text: str = None,
    ) -> str:
        '''
        Get the text object's string, masking if indicated.

        :param annot2mask: Mapping from annotation column (e.g., _num or
            _recsnum) to the replacement character(s) in the input text
            for masking already managed input.
        :param annot_df: Override annotations dataframe
        :param text: Override text
        :return: The (masked) text
        '''
        if annot2mask is None:
            return self.text
        # Apply the mask
        text_s = self.get_text_series(text=text)  # no padding
        if annot2mask is not None:
            annot_df = self.annotations.as_df
            text_s = self._apply_mask(text_s, annot2mask, annot_df)
        return ''.join(text_s)

    def get_text_series(
            self,
            pad_len: int = 0,
            text: str = None,
    ) -> pd.Series:
        '''
        Get the input text as a (padded) pandas series.
        :param pad_len: The number of spaces to pad both front ant back
        :param text: Override text
        :return: The (padded) pandas series of input characters.
        '''
        if text is None:
            text = self.text
        return pd.Series(list(
            ' '*pad_len + text + ' '*pad_len
        ))

    def get_annot_mask(
            self,
            annot_col: str,
            pad_len: int = 0,
            annot_df: pd.DataFrame = None,
            text: str = None,
    ) -> pd.Series:
        '''
        Get a True/False series for the input such that start to end positions
        for rows where the the annotation column is non-null and non-empty are
        True.
        :param annot_col: The annotation column identifying chars to mask
        :param pad_len: The number of characters to pad the mask with False
            values at both the front and back.
        :param annot_df: Override annotations dataframe
        :param text: Override text
        :return: A pandas Series where annotated input character positions
            are True and non-annotated positions are False.
        '''
        if annot_df is None:
            annot_df = self.annotations.as_df
        if text is None:
            text = self.text
        textlen = len(text)
        return self._get_annot_mask(
            annot_df,
            textlen,
            annot_col,
            pad_len=pad_len
        )

    @staticmethod
    def _get_annot_mask(
            annot_df: pd.DataFrame,
            textlen: int,
            annot_col: str,
            pad_len: int = 0,
    ) -> pd.Series:
        '''
        Get a True/False series for the input such that start to end positions
        for rows where the the annotation column is non-null and non-empty are
        True.

        :param annot_df: The annotations dataframe
        :param textlen: The length of the input text
        :param annot_col: The annotation column identifying chars to mask
        :param pad_len: The number of characters to pad the mask with False
            values at both the front and back.
        :return: A pandas Series where annotated input character positions
            are True and non-annotated positions are False.
        '''
        mask = None
        df = annot_df
        if annot_col in df.columns:
            df = df[np.logical_and(df[annot_col].notna(), df[annot_col] != '')]
            mask = pd.Series([False] * textlen)
            for _, row in df.iterrows():
                mask.loc[row['start_pos']+pad_len:row['end_pos']-1+pad_len] = True
        return mask
        
    def _apply_mask(
            self,
            text_s: pd.Series,
            annot2mask: Dict[str, str],
            annot_df: pd.DataFrame,
    ) -> str:
        if (
                len(text_s) > 0 and
                annot2mask is not None and
                annot_df is not None
        ):
            cols = set(
                annot_df.columns
            ).intersection(annot2mask.keys())
            if len(cols) > 0:
                for col in cols:
                    text_s = self._substitute(
                        text_s, col, annot2mask[col], annot_df,
                    )
        return text_s

    def _substitute(
            self,
            text_s: pd.Series,
            col: str,
            repl_mask: str,
            annot_df: pd.DataFrame,
    ) -> str:
        '''
        Substitute the "mask" char for "text" chars at "col"-annotated positions
        :param text_s: The text series to revise
        :param col: The annotation col identifying positions to mask
        :param repl_mask: The mask character to inject at annotated positions
        :param annot_df: The annotations dataframe
        :return: The masked text
        '''
        annot_mask = self._get_annot_mask(annot_df, len(text_s), col)
        text_s = text_s.mask(annot_mask, repl_mask)
        return text_s

    def add_annotations(self, annotations: Annotations):
        '''
        Add the annotations to this instance.

        :param annotations: The annotations to add.
        '''
        if annotations is not None and not annotations.is_empty():
            df = annotations.df
            if self._annots is None:
                self._annots = annotations
            elif self._annots.is_empty():
                if df is not None:
                    self._annots.set_df(df.copy())
            elif df is not None:
                self._annots.add_df(df)


class Annotator(ABC):
    '''
    Class for annotating text
    '''
    def __init__(
            self,
            name: str,
    ):
        '''
        :param name: The name of this annotator
        '''
        self.name = name

    @abstractmethod
    def annotate_input(
            self,
            text_obj: AnnotatedText,
            **kwargs,
    ) -> Annotations:
        '''
        Annotate this instance's text, additively updating its annotations.
        
        :param text_obj: The text object to annotate
        :return: The annotations added
        '''
        raise NotImplementedError


class BasicAnnotator(Annotator):
    '''
    Class for extracting basic (possibly multi -level or -part) entities.
    '''

    def annotate_input(
            self,
            text_obj: AnnotatedText,
            **kwargs,
    ) -> Annotations:
        '''
        Annotate the text obj, additively updating the annotations
        
        :param text: The text to annotate
        :return: The annotations added to the text
        '''
        # Get new annotation with just the syntax
        annots = self.annotate_text(text_obj.text)

        # Add syntactic annotations only as a bookmark
        text_obj.annotations.add_df(annots.as_df)

        return annots

    @abstractmethod
    def annotate_text(self, text_str: str) -> Annotations:
        '''
        Build annotations for the text string.
        :param text_str: The text string to annotate
        :return: Annotations for the text
        '''
        raise NotImplementedError


#TODO: remove this if unused -- stanza_annotator isa Authority -vs- stanza_annotator isa SyntacticParser
class SyntacticParser(BasicAnnotator):
    '''
    Class for creating syntactic annotations for an input.
    '''

    def annotate_input(
            self,
            text_obj: AnnotatedText,
            **kwargs,
    ) -> Annotations:
        '''
        Annotate the text, additively updating the annotations
        
        :param text: The text to annotate
        :return: The annotations added to the text
        '''
        # Get new annotation with just the syntax
        annots = self.annotate_text(text_obj.text)

        # Add syntactic annotations only as a bookmark
        text_obj.bookmarks[self.name] = annots.as_df

        return annots


class EntityAnnotator(BasicAnnotator):
    '''
    Class for extracting single (possibly multi-level or -part) entities.
    '''

    def __init__(
            self,
            name: str,
            mask_char: str = ' ',
    ):
        '''
        :param name: The name of this annotator
        :param mask_char: The character to use to mask out previously annotated
            spans of this annotator's text.
        '''
        super().__init__(name)
        self.mask_char = mask_char

    @property
    @abstractmethod
    def annotation_cols(self) -> Set[str]:
        '''
        Report the (final group or record) annotation columns that are filled
        by this annotator when its entities are annotated.
        '''
        raise NotImplementedError

    @abstractmethod
    def mark_records(
            self,
            annotations: Annotations,
            largest_only: bool = True
    ):
        '''
        Collect and mark annotation records.

        :param annotations: The annotations
        :param largest_only: True to only mark (keep) the largest records.
        '''
        raise NotImplementedError

    @abstractmethod
    def validate_records(
            self,
            annotations: Annotations,
    ):
        '''
        Validate annotated records.

        :param annotations: The annotations
        '''
        raise NotImplementedError

    @abstractmethod
    def compose_groups(
            self,
            annotations: Annotations
    ) -> Annotations:
        '''
        Compose annotation rows into groups.
        :param annotations: The annotations
        :return: The composed annotations
        '''
        raise NotImplementedError
        
    def annotate_input(
            self,
            text_obj: AnnotatedText,
            annot_mask_cols: Set[str] = None,
            merge_strategies: Dict[str, MergeStrategy] = None,
            largest_only: bool = True,
            **kwargs
    ) -> Annotations:
        '''
        Annotate the text object (optionally) after masking out previously
        annotated spans, additively updating the annotations in the text
        object.
        
        :param text_obj: The text object to annotate
        :param annot_mask_cols: The (possible) previous annotations whose
            spans to ignore in the text
        :param merge_strategies: A dictionary of each input annotation bookmark
            tag mapped to a merge strategy for merging this annotator's
            annotations with the bookmarked dataframe. This is useful, for
            example, when merging syntactic information to refine ambiguities.
        :param largest_only: True to only mark largest records.
        :return: The annotations added to the text object
        '''
        annot2mask = None if annot_mask_cols is None else { #TODO: Use this?!
            col: self.mask_char
            for col in annot_mask_cols
        }
        
        annots = self.annotate_text(text_obj.text)
        if annots is None:
            return annots

        if merge_strategies is not None:
            bookmarks = text_obj.bookmarks
            if bookmarks is not None and len(bookmarks) > 0:
                for tag, merge_strategy in merge_strategies.items():
                    if tag in bookmarks:
                        text_obj.bookmarks[f'{self.name}.pre-merge:{tag}'] = annots.df
                        annots.add_df(bookmarks[tag])
                        annots = merge(annots, merge_strategy)

        annots = self.compose_groups(annots)

        self.mark_records(annots, largest_only=largest_only)
        # NOTE: don't pass "text" here because it may be masked
        self.validate_records(annots)
        text_obj.annotations.add_df(annots.df)
        return annots

    @property
    @abstractmethod
    def highlight_fieldstyles(self) -> Dict[str, Dict[str, Dict[str, str]]]:
        '''
        Get highlight field styles for this annotator's annotations of the form:
        {
            <field_col>: {
                <field_value>: {
                    <css-attr>: <css-value>
                }
            }
        }
        For css-attr's like 'background-color', 'foreground-color', etc.
        '''
        raise NotImplementedError


class HtmlHighlighter:
    '''
    Helper class to add HTML markup for highlighting spans of text.
    '''
    def __init__(
            self,
            field2style: Dict[str, Dict[str, str]],
            tooltip_class='tooltip',
            tooltiptext_class='tooltiptext',
    ):
        '''
        :param field2style: The annotation column to highlight with its
            associated style, for example:
                {
                    'car_model_field': {
                        'year': {'background-color': 'lightyellow'},
                        'make': {'background-color': 'lightgreen'},
                        'model': {'background-color': 'cyan'},
                        'style': {'background-color': 'magenta'},
                    },
                }
        :param tooltip_class: The css tooltip class
        :param tooltiptext_class: The css tooltiptext class
        '''
        self.field2style = field2style
        self.tooltip_class = tooltip_class
        self.tooltiptext_class = tooltiptext_class

    def highlight(
            self,
            text_obj: AnnotatedText,
    ) -> str:
        '''
        Return an html string with the given fields (annotation columns)
        highlighted with the associated styles.
        :param text_obj: The annotated text to markup
        '''
        result = ['<p>']
        anns = text_obj.annotations
        an_df = anns.df
        for field, styles in self.field2style.items():
            # NOTE: the following line relies on an_df already being sorted
            df = an_df[an_df[field].isin(styles)]
            cur_pos = 0
            for _loc, row in df.iterrows():
                enttype = row[field]
                style = styles[enttype]
                style_str = ' '.join([
                    f'{key}: {value};'
                    for key, value in style.items()
                ])
                start_pos = row[anns.metadata.start_pos_col]
                if start_pos > cur_pos:
                    result.append(text_obj.text[cur_pos:start_pos])
                end_pos = row[anns.metadata.end_pos_col]
                result.append(
                    f'<mark class="{self.tooltip_class}" style="{style_str}">'
                )
                result.append(text_obj.text[start_pos:end_pos])
                result.append(
                    f'<span class="{self.tooltiptext_class}">{enttype}</span>'
                )
                result.append('</mark>')
                cur_pos = end_pos
        result.append('</p>')
        return '\n'.join(result)


class AnnotatorKernel(ABC):
    '''
    Class for encapsulating core annotation logic for multiple annotators
    '''

    @property
    @abstractmethod
    def annotators(self) -> List[EntityAnnotator]:
        ''' Get the entity annotators '''
        raise NotImplementedError

    @abstractmethod
    def annotate_input(self, text_obj: AnnotatedText) -> Annotations:
        ''' Execute all annotations on the text_obj '''
        raise NotImplementedError


class CompoundAnnotator(Annotator):
    '''
    Class to apply a series of annotators through an AnnotatorKernel
    '''

    def __init__(
            self,
            kernel: AnnotatorKernel,
            name: str = 'entity',
    ):
        '''
        Initialize with the annotators and this extractor's name.

        :param kernel: The annotations kernel to use
        :param name: The name of this information extractor to be the
            annotations base column name for <name>_num and <name>_recsnum
        '''
        super().__init__(name=name)
        self.kernel = kernel

    def annotate_input(
            self,
            text_obj: AnnotatedText,
            reset: bool = True,
            **kwargs,
    ) -> Annotations:
        '''
        Annotate the text.

        :param text_obj: The AnnotatedText object to annotate.
        :param reset: When True, reset and rebuild any existing annotations
        :return: The annotations added to the text_obj
        '''
        if reset:
            text_obj.annotations.clear()
        annots = self.kernel.annotate_input(text_obj)
        return annots

    def get_html_highlighted_text(
            self,
            text_obj: AnnotatedText,
            annotator_names: List[str] = None,
    ) -> str:
        '''
        Get html-hilighted text for the identified input's annotations
        from the given annotators (or all).

        :param text_obj: The input text to highlight
        :param annotator_names: The subset of annotators to highlight.
        '''
        if annotator_names is None:
            annotator_names = [ann.name for ann in self.kernel.annotators]
        hfs = {
            ann.name: ann.highlight_fieldstyles
            for ann in self.kernel.annotators
            if ann.name in annotator_names
        }
        hh = HtmlHighlighter(hfs)
        return hh.highlight(text_obj)



'''
General data utilities for working with 3rd party resources
'''
import nltk
import os


# NLTK Wordnet Resources
NLTK_RESOURCES_PATH = 'opt/nltk_resources'
NLTK_RESOURCES = {
    'wordnet': os.path.join('corpora', 'wordnet.zip'),
    'omw-1.4': os.path.join('corpora', 'omw-1.4.zip'),
    'wordnet_ic': os.path.join('corpora', 'wordnet_ic.zip')
}


# Runtime variables
_DATADIR = None
_NLTK_RESOURCES_DIR = None
_NLTK_WN = None


def active_datadir():
    '''
    Get the active data directory as the first that exists from:
        * the DATADIR environment variable
        * the HOME/data directory
        * the /data directory

    NOTE: If an active DATADIR must exist to get a non-None result.

    :return: The active data directory or None
    '''
    global _DATADIR  # pylint: disable-msg=W0603
    if _DATADIR is None:
        _DATADIR = os.environ.get(
            'DATADIR',
            os.environ.get('HOME', '') + '/data'
        )
        if not os.path.exists(_DATADIR) and os.path.exists('/data'):
            _DATADIR = '/data'
    return _DATADIR


def download_nltk_resources(resources, resources_dir=None, verbose=False, downloader=nltk.download):
    '''
    Download the nltk resources that don't yet exist.
    :resources_dir: The resources (root) directory to download to, or
        the default if None.
    :resources: A dictionary of resource names to download mapped to
        the relative (to the resoures dir) path of the resource
    :verbose: True to print status
    '''
    if resources is not None:
        if resources_dir is None:
            resources_dir = get_nltk_resources_dir()
        for resource, relpath in resources.items():
            respath = os.path.join(resources_dir, relpath)
            if not os.path.exists(respath):
                if verbose:
                    print(f'NOTE: {respath} does not exist. Downloading...')
                downloader(resource, download_dir=resources_dir)


def get_nltk_resources_dir(resources=None, verbose=False, downloader=nltk.download):
    '''
    Get the NLTK resources directory, optionally downloading resources, from:
        * the NLTK_DATA environment variable
        * the active_datadir()/NLTK_RESOURCES_PATH

    NOTE: If an active DATADIR must exist to get a non-None result.

    :param resources: A dictionary identifying the resources and relative (to
        the NLTK resources directory) paths to ensure are downloaded
    :verbose: True to print status
    :return: The NLTK resources directory path or None
    '''
    global _NLTK_RESOURCES_DIR  # pylint: disable-msg=W0603
    if _NLTK_RESOURCES_DIR is None:
        _NLTK_RESOURCES_DIR = os.environ.get('NLTK_DATA', None)
        if _NLTK_RESOURCES_DIR is None:
            datadir = active_datadir()
            if datadir is not None:
                resdir = os.path.join(datadir, NLTK_RESOURCES_PATH)
                _NLTK_RESOURCES_DIR = resdir
                os.environ['NLTK_DATA'] = resdir
                nltk.data.path.append(resdir)
                if resources is not None:
                    download_nltk_resources(
                        resources, resources_dir=resdir, verbose=verbose,
                        downloader=downloader,
                    )
    return _NLTK_RESOURCES_DIR


def get_nltk_wordnet(downloader=nltk.download):
    '''
    Get a handle on NLTK's wordnet object, ensuring resources have
    been downloaded.
    :return: nltk.corpus.wordnet
    '''
    # Make sure resources have been downloaded
    global _NLTK_WN  # pylint: disable-msg=W0603
    if _NLTK_WN is None:
        if get_nltk_resources_dir(resources=NLTK_RESOURCES, downloader=downloader) is not None:
            from nltk.corpus import wordnet as wn
            _NLTK_WN = wn
    return _NLTK_WN



import gzip
import io
import json_stream.requests
import os
import pandas as pd
import re
import requests
import dataknobs.structures.tree as dk_tree
import dataknobs.utils.file_utils as dk_futils
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Any, Callable, Dict, List, Set, Tuple, Union


ELT_IDX_RE = re.compile(r'^(.*)\[(.*)\]$')
FLATTEN_IDX_RE = re.compile(r'\[(\d+)\]')
URL_RE = re.compile(r'^\s*https?://.*$', flags=re.IGNORECASE)
TIMEOUT = 10  # 10 seconds


def stream_json_data(
        json_data: str,
        visitor_fn: Callable[Any, str],
        timeout: int = TIMEOUT,
):
    '''
    Stream the json data, calling the visitor_fn at each value.
    :param json_data: The json data (url, file_path, or str)
    :param visitor_fn: The visitor_fn(item, path) to call, where
        item is each json item's value and path is the tuple of
        elements identifying the path to the item.
    :param timeout: The requests timeout (in seconds)
    '''
    if os.path.exists(json_data):
        if dk_futils.is_gzip_file(json_data):
            with gzip.open(json_data, 'rt', encoding='utf-8') as f:
                json_stream.visit(f, visitor_fn)
        else:
            with open(json_data, 'r', encoding='utf-8') as f:
                json_stream.visit(f, visitor_fn)
    elif json_data.startswith('http'):
        with requests.get(json_data, stream=True, timeout=timeout) as response:
            json_stream.requests.visit(response, visitor_fn)
    elif isinstance(json_data, str):
        f = io.StringIO(json_data)
        json_stream.visit(f, visitor_fn)


def build_jq_path(path: Tuple[Any], keep_list_idxs=True) -> str:
    '''
    Build a jq path string from the json_stream path tuple.
    :param path: A tuple with json_stream path components.
    :param keep_list_idxs: True to keep the exact path index values;
        False to emit a generic "[]" for the list.
    '''
    jq_path = ''
    for elt in path:
        if isinstance(elt, int):
            jq_path += f'[{elt}]' if keep_list_idxs and elt>=0 else '[]'
        else:
            jq_path += f'.{elt}'
    return jq_path


def build_path_tuple(jq_path: str, any_list_idx: int = -1) -> Tuple:
    '''
    Build a json_stream tuple path from a jq_path (reverse of build_jq_path).
    :param jq_path: The jq_path whose values to extract.
    :param any_list_idx: The index value to give for a generic list
    :return: The json_stream tuple form of the path
    '''
    path = list()
    for part in jq_path.split('.'):
        if part == '':
            continue
        m = ELT_IDX_RE.match(part)
        if m:
            path.append(m.group(1))
            idx = m.group(2)
            idxval = any_list_idx if idx == '' else int(idx)
            path.append(idxval)
        else:
            path.append(part)
    return tuple(path)


def stream_jq_paths(
        json_data: str,
        output_stream,
        line_builder_fn: Callable[[str, Any], str] = (
            lambda jq_path,item: f'{jq_path}\t{item}'
        ),
        keep_list_idxs: bool = True,
        timeout: int = TIMEOUT,
):
    '''
    Write built lines from (jq_path, item) tuples from the json_data.
    :param json_data: The json_data to copy
    :param output_stream: The output stream to write lines to
    :param line_builder_fn: The function for turning record tuples into a line
        string.
    :param keep_list_idxs: True to keep the exact path index values;
        False to emit a generic "[]" for the list.
    :param timeout: The requests timeout (in seconds)
    '''
    def visitor(item, path):
        jq_path = build_jq_path(path, keep_list_idxs=keep_list_idxs)
        line = line_builder_fn(jq_path, item)
        print(line, file=output_stream)

    stream_json_data(json_data, visitor, timeout=timeout)


def squash_data(
        builder_fn: Callable[str, Any],
        json_data: str,
        prune_at: List[Union[str, Tuple[str, int]]] = None,
        timeout: int = TIMEOUT,
):
    '''
    Squash the json_data, optionally pruning branches by id'd path elements.
    Where "squashed" means the paths are compressed to a single level
    with complex jq keys.

    Each path to prune is identified by:
        * (path_element_name:str, path_index:int)
            -- or paths where path[path_index] == path_element_name
        * path_element_name:str or (path_element_name, None)
            -- or any path element whose name is path_element_name
        * path_index:int or (None, path_index)
            -- or any path at path_index depth

    :param builder_fn: fn(jq_path, item) for building a result
    :param json_data: The json_data to copy
    :param prune_at: The names and optional path indexes of path elements
        to ignore
    :param timeout: The requests timeout (in seconds)
    '''
    raw_depths = set()
    raw_elts = set()
    elt_depths = dict()
    depth_elts = defaultdict(set)

    def decode_item(item):
        if item is not None:
            if isinstance(item, str):
                raw_elts.add(item)
            elif isinstance(item, int):
                raw_depths.add(item)
            elif isinstance(item, tuple) and len(item) == 2:
                elt = item[0]
                depth = item[1]
                if elt is None:
                    if depth is not None:
                        raw_depths.add(depth)
                else:
                    if depth is not None:
                        depth_elts[depth].add(elt)
                        elt_depths[elt] = depth
                    else:
                        raw_elts.add(elt)
            elif isinstance(item, list) or isinstance(item, tuple):
                for i in item:
                    decode_item(i)

    decode_item(prune_at)
    has_raw_depths = len(raw_depths) > 0
    has_raw_elts = len(raw_elts) > 0
    has_elts = len(elt_depths) > 0
    has_depth_elts = len(depth_elts) > 0
    do_prune = has_raw_depths or has_raw_elts or has_elts or has_depth_elts

    def visitor(item, path):
        if do_prune:
            cur_depth = len(path)
            if has_raw_depths and cur_depth in raw_depths:
                return
            if has_raw_elts:
                if len(raw_elts.intersection(path)) > 0:
                    return
            cur_elt = path[-1]
            if has_elts and cur_elt in elt_depths:
                if cur_depth == elt_depths[cur_elt]:
                    return
            if has_depth_elts:
                for depth, elts in depth_elts.items():
                    if depth < cur_depth and path[depth] in elts:
                        return
        # Add squashed element
        jq_path = build_jq_path(path, keep_list_idxs=True)
        builder_fn(jq_path, item)

    stream_json_data(json_data, visitor, timeout=timeout)


def collect_squashed(
        jdata: str,
        prune_at: List[Union[str, Tuple[str, int]]] = None,
        timeout: int = TIMEOUT,
        result: Dict = None,
) -> Dict:
    '''
    Collected squashed data in a dictionary
    :param json_data: The json_data to copy
    :param prune_at: The names and optional path indexes of path elements
        to ignore
    :param timeout: The requests timeout (in seconds)
    :param result: (optional) The dictionary in which to collect items
    '''
    if result is None:
        result = dict()
    def collector_fn(jq_path, item):
        result[jq_path] = item
    squash_data(
        collector_fn,
        jdata,
        prune_at=prune_at,
        timeout=timeout,
    )
    return result


def indexing_format_fn(jq_path: str, item: Any) -> str:
    '''
    A formatting function for writing a csv with the columns:
        item (value) \t field \t flat_jq \t idxs

    Where
        * value -- is the item value
        * field -- is the last flat_jq path element
        * flat_jq -- is the jq_path with all indexes flattened "[.*]" -> "[]"
                     up to the last path element
        * idxs -- is a comma-delimitted string with the indices
    '''
    idxs = ', '.join(FLATTEN_IDX_RE.findall(jq_path))
    flat_jq = FLATTEN_IDX_RE.sub('[]', jq_path)
    dotpos = flat_jq.rindex('.')
    field = flat_jq[dotpos+1:]
    flat_jq = flat_jq[:dotpos]
    return f'{item}\t{field}\t{flat_jq}\t{idxs}'


def indexing_format_splitter(fileline: str) -> Tuple[str, str, str, str]:
    '''
    Reversal of the indexing_format_fn to extract:
        (value, field, flat_jq, idxs)
    Where
        * value -- is the item value
        * field -- is the last flat_jq path element
        * flat_jq -- is the jq_path with all indexes flattened "[.*]" -> "[]"
                     up to the last path element
        * idxs -- is a comma-delimitted string with the indices
    '''
    line = fileline.strip()
    value = None
    field = None
    flat_jq = None
    idxs = None
    if line:
        parts = fileline.split('\t')
        value = parts[0]
        if len(parts) > 1:
            field = parts[1] 
            if len(parts) > 2:
                flat_jq = parts[2] 
                if len(parts) > 3:
                    idxs = parts[3] 
    return (value, field, flat_jq, idxs)


def write_squashed(
        dest_file: str,
        jdata: str,
        prune_at: List[Union[str, Tuple[str, int]]] = None,
        timeout: int = TIMEOUT,
        format_fn: Callable[[str, Any], str] = lambda jq_path, item: f'{jq_path}\t{item}',
):
    '''
    Write squashed data to the file.
    :param json_data: The json_data to copy
    :param prune_at: The names and optional path indexes of path elements
        to ignore
    :param timeout: The requests timeout (in seconds)
    :param format_fn: A function for formatting each output line. Default is
        tab-delimitted: jq_path <tab> item.
    '''
    needs_close = False
    if isinstance(dest_file, str):
        f = open(dest_file, 'w', encoding='utf-8')
        needs_close = True
    else:
        f = dest_file
    squash_data(
        lambda jq_path, item: print(format_fn(jq_path, item), file=f),
        jdata,
        prune_at=prune_at,
        timeout=timeout,
    )
    if needs_close:
        f.close()


def path_to_dict(path: Union[tuple, str], value: Any, result=None) -> dict:
    '''
    Convert the jq_path (if a string) or path (if a tuple) to a dict.
    :param path: The path to convert
    :param value: The path's value
    :param result: The dictionary to add the result to
    :return: the result
    '''
    if result is None:
        result = dict()
    if isinstance(path, str):
        path = build_path_tuple(path, any_list_idx=-1)

    def do_it(cur_dict, path, path_idx, pathlen, value):
        if path_idx >= pathlen:
            return
        path_elt = path[path_idx]
        path_idx += 1
        list_idx = None
        if path_idx < pathlen:
            next_elt = path[path_idx]
            if isinstance(next_elt, int):
                list_idx = next_elt
                path_idx += 1

        if path_elt in cur_dict:
            if list_idx is None:
                if path_idx < pathlen:
                    cur_dict = cur_dict[path_elt]
                else:
                    cur_dict[path_elt] = value
            else:
                cur_list = cur_dict[path_elt]
                if len(cur_list) <= list_idx:
                    if path_idx < pathlen:
                        cur_dict = dict()
                        # simplifying assumption: idxs are in consecutive order fm 0
                        cur_list.append(cur_dict)
                    else:
                        cur_list.append(value)
                else:
                    if path_idx < pathlen:
                        cur_dict = cur_list[-1]
                    else:
                        cur_list.append(value)
        else:
            if list_idx is None:
                if path_idx < pathlen:
                    elt_dict = dict()
                    cur_dict[path_elt] = elt_dict
                    cur_dict = elt_dict
                else:
                    cur_dict[path_elt] = value
            else:
                cur_list = list()
                cur_dict[path_elt] = cur_list
                if path_idx < pathlen:
                    cur_dict = dict()
                    cur_list.append(cur_dict)
                else:
                    cur_list.append(value)
        # recurse to keep moving along the path
        do_it(cur_dict, path, path_idx, pathlen, value)


    do_it(result, path, 0, len(path), value)

    return result


def explode(squashed: Dict) -> Dict:
    '''
    Explode a "squashed" json with jq paths as keys.
    '''
    result = dict()
    for jq_path, value in squashed.items():
        path_to_dict(jq_path, value, result)
    return result


class ValuePath:
    '''
    Structure to hold (compressed) information about the jq_path indices
    leading to a unique value.

    Essentially, this the tree of path indices leading to the value.
    '''
    def __init__(self, jq_path: str, value: Any):
        '''
        :param jq_path: The jq_path (key)
        :param value: The path's value
        '''
        self.jq_path = jq_path
        self.value = value
        self._indices = dk_tree.Tree(0).as_string()  # root data will hold total path count

    @property
    def indices(self) -> dk_tree:
        return dk_tree.build_tree_from_string(self._indices)

    def add(self, path: tuple):
        '''
        Add the path (with the samem structure as jq_path) to the tree.
        '''
        root = self.indices
        node = root
        node.data = int(node.data) + 1  # keep track of total
        if path is None:
            self._indices = root.as_string()
            return
        for elt in path:
            if isinstance(elt, int):
                found = False
                if node.has_children():
                    # simplifying assumption: idxs are in consecutive order fm 0
                    child = node.children[-1]
                    if str(elt) == child.data:
                        node = child
                        found = True
                if not found:
                    node = node.add_child(elt)
        self._indices = root.as_string()

    @property
    def path_count(self) -> int:
        ''' Get the number of jq_paths to the value. '''
        return int(self.indices.data)

    def path_generator(self, result_type='jq_path'):
        '''
        Generate value paths.
        :param result_type: 'jq_path', 'path', or 'idx'
            'jq_path' to generate jq_path strings;
            'path' to generate path tuples
            'idx' to generate index tuples.
        '''
        path = build_path_tuple(self.jq_path, any_list_idx=-1)
        for node in self.indices.collect_terminal_nodes():
            node_path = node.get_path()
            node_idx = 0
            gen_path = list()
            for elt in path:
                if isinstance(elt, int):
                    gen_path.append(int(node_path[node_idx].data))
                    node_idx += 1
                elif result_type != 'idx':
                    gen_path.append(elt)
            if result_type == 'jq_path':
                yield build_jq_path(gen_path, keep_list_idxs=True)
            else:
                yield gen_path


class ValuesIndex:
    '''
    An index of unique values by jpath and (optionally) an index of values to
    their paths for each jpath.

    The values to paths index is compressed such that each unique value mapped
    to the tree of path indices leading to it, if path information is available.
    '''
    def __init__(self):
        self.path_values = dict()  # Dict[jq_path, Dict[value, ValuePath]]

    def add(self, value: Any, jq_path: str, path: tuple = None):
        if jq_path in self.path_values:
            value_paths = self.path_values[jq_path]
        else:
            value_paths = dict()
            self.path_values[jq_path] = value_paths

        if value == []:
            value = '_EMPTY_LIST_'
        elif value == {}:
            value = '_EMPTY_DICT_'
            
        if value in value_paths:
            value_path = value_paths[value]
        else:
            value_path = ValuePath(jq_path, value)
            value_paths[value] = value_path

        value_path.add(path)

    def has_jqpath(self, jq_path: str) -> bool:
        ''' Determine whether there are any values for jq_path '''
        return jq_path in self.path_values

    def get_values(self, jq_path: str) -> Set[Any]:
        ''' Get the set of values for jq_path '''
        return set(self.path_values.get(jq_path, {}).keys())

    def num_values(self, jq_path: str) -> int:
        ''' Get the number of values for jq_path '''
        return len(self.path_values.get(jq_path, {}))


class JsonSchema:
    '''
    Container for a schema view of a json object of the form:

    {
        <jq_path>: {
            <value_type>: <value_count>,
        }
        ...,
    }

    Where,
        * value_type is the type of value at the path (e.g., int, float,
            str, etc.)
        * value_count is the number of types the value_type occurs in an object.
    '''

    def __init__(
            self,
            schema: Dict[str, Any] = None,
            values: ValuesIndex = None,
            values_limit: int = 0,  # max number of unique values to keep
    ):
        '''
        :param schema: Reconstruct instance with a known schema
        :param values: Prime with existing values
        :param values_limit: Stop counting unique values after reaching this
            limit, or don't stop counting if 0.
        '''
        self.schema = schema if schema is not None else dict()
        self.values = ValuesIndex() if values is None else values
        self._values_limit = values_limit
        self._df = None

    def add_path(
            self,
            jq_path: str,
            value_type: str,
            value: Any = None,
            path: tuple = None,
    ):
        '''
        Add an instance of the jq_path/value_type
        :param jq_path: The "key" path for grouping/squashing values
        :param value_type: The type of value with this path
        :param value: (optional) The value for tracking unique values (if not
            None)
        :param path: (optional) The path tuple for inverting paths to uniques
            (if not None and if value is also not None)
        '''
        if jq_path not in self.schema:
            self.schema[jq_path] = {value_type: 1}
        else:
            if value_type not in self.schema[jq_path]:
                self.schema[jq_path][value_type] = 1
            else:
                self.schema[jq_path][value_type] += 1
        if value is not None:
            if (
                    self._values_limit == 0 or
                    self.values.num_values(jq_path) < self._values_limit
            ):
                self.values.add(value, jq_path, path=path)
        self._df = None
        
    @property
    def df(self) -> pd.DataFrame:
        '''
        Get schema information as a DataFrame with columns:

            jq_path    value_type    value_count
        '''
        if self._df is None:
            self._df = self._build_df()
        return self._df

    def _build_df(self) -> pd.DataFrame:
        '''
        Get schema information as a DataFrame with columns:

            jq_path    value_type    value_count    [unique_count]
        '''
        data = list()
        has_value = False
        for k1, v1 in self.schema.items():  # jq_path -> [value_type -> value_count]
            for k2, v2 in v1.items():  # value_type -> value_count
                if self.values.has_jqpath(k1):
                    row = (k1, k2, v2, self.values.num_values(k1))
                    has_value = True
                else:
                    row = (k1, k2, v2)
                data.append(row)
        columns = ['jq_path', 'value_type', 'value_count']
        if has_value:
            columns.append('unique_count')
        return pd.DataFrame(data, columns=columns)

    def extract_values(
            self,
            jq_path: str,
            json_data: str,
            unique: bool = True,
            timeout: int = TIMEOUT,
    ) -> Union[List[Any], Set[Any]]:
        '''
        Extract values from the json_data's jq_path.
        :param jq_path: The jq_path whose values to extract.
        :param json_data: The json data (url, file_path, or str)
        :param unique: True to collect only unique values
        :param timeout: The requests timeout (in seconds)
        :return: The list (or set if unique) of values.
        '''
        keep_list_idxs = False if '[]' in jq_path else False
        sresult = set()
        lresult = list()
        def visitor(item, path):
            cur_jq_path = build_jq_path(path, keep_list_idxs=keep_list_idxs)
            if jq_path == cur_jq_path:
                if unique:
                    sresult.add(item)
                else:
                    lresult.append(item)
        stream_json_data(json_data, visitor, timeout=timeout)
        return sresult if unique else lresult


class JsonSchemaBuilder:
    '''
    Create a schema view of a json object.
    '''

    def __init__(
            self,
            json_data: str,
            value_typer: Callable[[Any], str] = None,
            keep_unique_values: bool = False,
            invert_uniques: bool = False,
            keep_list_idxs: bool = False,
            timeout: int = TIMEOUT,
            empty_dict_type: str = '_EMPTY_DICT_',
            empty_list_type: str = '_EMPTY_LIST_',
            unk_value_type: str = '_UNKNOWN_',
            int_value_type: str = 'int',
            float_value_type: str = 'float',
            str_value_type: str = 'str',
            url_value_type: str = 'URL',
            on_add: Callable[[str], bool] = None,
    ):
        '''
        :param json_data: The json data (url, file_path, or str)
        :param value_typer: A fn(atomic_value) that returns the type of the
            value to override the default typing of "int", "float", and "str"
        :param keep_unique_values: True to keep unique values for each path or
            an integer for a maximum number of unique values to keep
        :param invert_uniques: True to keep value paths for unique values
        :param keep_list_idxs: True to keep the list indexes in the dictionary
            paths. When False, all list indexes will be generalized to "[]".
        :param timeout: The requests timeout (in seconds)
        :param empty_dict_type: The type of an empty dictionary
        :param empty_list_type: The type of an empty list
        :param unk_value_type: The type of an unknown/unclassified value
        :param int_value_type: The type of an int
        :param float_value_type: The type of a float
        :param str_value_type: The type of a string
        :param url_value_type: The value type of a URL if not None
        :param on_add: A fn(jq_path) called just before adding each jq_path
            that returns True to add the path and False to skip adding it.
            When None, all paths are added.
        '''
        self.json_data = json_data
        self.value_typer = value_typer
        self.keep_uniques = keep_unique_values
        self.values_limit = (
            0
            if isinstance(keep_unique_values, bool)
            else int(keep_unique_values)
        )
        self.invert_uniques = invert_uniques
        self.keep_list_idxs = keep_list_idxs
        self.timeout = timeout
        self.empty_dict_type = empty_dict_type
        self.empty_list_type = empty_list_type
        self.unk_value_type = unk_value_type
        self.int_value_type = int_value_type
        self.float_value_type = float_value_type
        self.str_value_type = str_value_type
        self.url_value_type = url_value_type
        self._on_add = on_add
        self._schema = JsonSchema(values_limit = self.values_limit)
        self._built_schema = False

    @property
    def schema(self) -> JsonSchema:
        ''' Get the schema for the json data '''
        if not self._built_schema:
            self._built_schema = self._build_schema()
        return self._schema

    @property
    def partial_schema(self) -> JsonSchema:
        ''' Get the current, possibly incomplete, schema '''
        return self._schema

    def _build_schema(self) -> JsonSchema:
        '''
        Stream the json data and build the schema.
        '''
        def visitor(item, path):
            self._visit_item(self._schema, item, path)
        stream_json_data(self.json_data, visitor, timeout=self.timeout)
        return True

    def _visit_item(self, schema: JsonSchema, item: Any, path: Tuple):
        '''
        The visitor function for processing items and paths while streaming
        the json data.
        :param schema: The schema being built
        :param item: The next json item (value) encountered
        :param path: The path of the item as a tuple of the form:
            (<key1>, <list1-idx>, <key2>, ...)
        '''
        jq_path = build_jq_path(path, keep_list_idxs = self.keep_list_idxs)
        value_type = None
        if isinstance(item, dict):
            value_type = self.empty_dict_type
        elif isinstance(item, list):
            value_type = self.empty_list_type
        elif self.value_typer is not None:
            value_type = self.value_typer(item)
        if value_type is None:
            if isinstance(item, str):
                if self.url_value_type and URL_RE.match(item):
                    value_type = self.url_value_type
                else:
                    value_type = self.str_value_type
            elif isinstance(item, float):
                value_type = self.float_value_type
            elif isinstance(item, int):
                value_type = self.int_value_type
            else:
                value_type = self.unk_value_type
        do_add = True
        if self._on_add is not None:
            do_add = self._on_add(jq_path)
        if do_add:
            schema.add_path(
                jq_path, value_type,
                value=(item if self.keep_uniques else None),
                path=(path if self.invert_uniques else None),
            )


class Path:
    '''
    Container for a path.
    '''
    def __init__(self, jq_path: str, item: Any, line_num: int = -1):
        '''
        :param jq_path: A fully-qualified indexed path.
        :param item: The path's item (value)
        '''
        self.jq_path = jq_path
        self.item = item
        self.line_num = line_num
        self._path_elts = None  # jq_path.split('.')
        self._len = None  # Number of path elements

    def __repr__(self) -> str:
        lnstr = f'{self.line_num}: ' if self.line_num >= 0 else ''
        return f'{lnstr}{self.jq_path}: {self.item}'

    def __key(self):
        return (self.jq_path, self.item) if self.line_num < 0 else self.line_num

    def __lt__(self, other: 'Path') -> bool:
        if self.line_num < 0 or other.line_num < 0:
            return self.jq_path < other.jq_path
        else:
            return self.line_num < other.line_num

    def __hash__(self) -> int:
        return hash(self.__key())

    def __eq__(self, other):
        if isinstance(other, Path):
            return self.__key() == other.__key()
        return NotImplemented

    @property
    def path_elts(self) -> List[str]:
        ''' Get this path's (index-qualified) elements '''
        if self._path_elts is None:
            self._path_elts = self.jq_path.split('.')
        return self._path_elts

    @property
    def size(self) -> int:
        ''' Get the number of path_elements in this path. '''
        return len(self.path_elts)


class GroupAcceptStrategy(ABC):
    '''
    Add a Path to a Group if it belongs.
    '''
    @abstractmethod
    def accept_path(
            self,
            path: Path,
            group: 'PathGroup',
            distribute: bool = False
    ) -> str:
        '''
        Determine whether the Path belongs in the group.
        :param path: The path to potentially add
        :param group: The group to which to add the path
        :param distribute: True if the path is proposed as a distributed path
            instead of as a main path.
        :return: 'main', 'distributed', or None if the path belongs as a main
            path, a distributed path, or not at all in the group
        '''
        raise NotImplementedError


class PathGroup:
    '''
    Container for a group of related paths.
    '''
    def __init__(
            self,
            accept_strategy: GroupAcceptStrategy,
            first_path: Path = None
    ):
        self._all_paths = None
        self.main_paths = None  # Set[Path]
        self.distributed_paths = None  # Set[Path]
        self.accept_strategy = accept_strategy
        if first_path is not None:
            self.accept(first_path, distribute=False)

    @property
    def num_main_paths(self) -> int:
        ''' Get the number of main paths in this group '''
        return len(self.main_paths) if self.main_paths is not None else 0

    @property
    def num_distributed_paths(self) -> int:
        ''' Get the number of distributed paths in this group '''
        return (
            len(self.distributed_paths)
            if self.distributed_paths is not None
            else 0
        )

    @property
    def size(self) -> int:
        ''' Get the total number of paths in this group. '''
        return self.num_main_paths + self.num_distributed_paths

    @property
    def paths(self) -> List[Path]:
        ''' Get all paths (both main and distributed) '''
        if self._all_paths is None:
            if self.main_paths is not None:
                self._all_paths = self.main_paths.copy()
                if self.distributed_paths is not None:
                    self._all_paths.update(self.distributed_paths)
                self._all_paths = sorted(self._all_paths)
            elif self.distributed_paths is not None:
                self._all_paths = sorted(self.distributed_paths)
        return self._all_paths

    def as_dict(self) -> Dict[str, str]:
        ''' Reconstruct the object from the paths '''
        d = dict()
        if self.paths is not None:
            for path in self.paths:
                path_to_dict(path.jq_path, path.item, result=d)
        return d

    def accept(self, path: Path, distribute: bool = False) -> bool:
        '''
        Add the path if it belongs in this group.
        :param path: The path to (potentially) add.
        :param distribute: True to propose the path as a distributed path
        :return: True if the path was accepted and added.
        '''
        added = False
        add_type = self.accept_strategy.accept_path(
            path, self, distribute=distribute
        )
        if add_type is not None:
            if add_type == 'main':
                if self.main_paths is None:
                    self.main_paths = {path}
                else:
                    self.main_paths.add(path)
            else:  # 'distributed'
                if self.distributed_paths is None:
                    self.distributed_paths = {path}
                else:
                    self.distributed_paths.add(path)
            added = True
            self._all_paths = None
        return added

    def incorporate_paths(self, group: 'PathGroup'):
        '''
        Incorporate (distribute) the group's appliccable paths into this group.
        '''
        for path in group.paths:
            self.accept(path, distribute=True)


class ArrayElementAcceptStrategy(GroupAcceptStrategy):
    '''
    Container for a consistent group of paths built around each array.
    '''
    def __init__(self, max_array_level: int = -1):
        '''
        :param max_array_level: -1 to ignore, 0 to force new record at the
            first (top) array level, 1 at the 2nd, etc.
        '''
        self.max_array_level = max_array_level
        self.ref_path = None

    def accept_path(
            self,
            path: Path,
            group: PathGroup,
            distribute: bool = False
    ):
        if distribute or not '[' in path.jq_path:
            return 'distributed'

        if group.num_main_paths == 0:
            # Accept first path with an array
            self.ref_path = path
            return 'main'
        else:
            if self.ref_path is None:
                self.ref_path = list(group.main_paths)[0]
            # All elements up through max_array_level must fully match
            cur_array_level = -1
            for idx in range(1, min(self.ref_path.size, path.size)):
                ref_elt = self.ref_path.path_elts[idx]
                path_elt = path.path_elts[idx]
                if ref_elt != path_elt:
                    return None
                elif ref_elt[-1] == ']':
                    cur_array_level += 1
                    if cur_array_level >= self.max_array_level:
                        break
        return 'main'


class PathSorter:
    '''
    Container for sorting paths belonging together into groups.
    '''
    def __init__(
            self,
            accept_strategy: GroupAcceptStrategy,
            group_size: int = 0,
            max_groups: int = 0,
    ):
        '''
        :param accept_strategy: The group accept strategy to use
        :param group_size: A size constraint/expectation for groups such that
           any group not reaching this size (if > 0) is silently "dropped".
        :param max_groups: The maximum number of groups to keep in memory where
            all groups are kept if <= 0
        '''
        self.accept_strategy = accept_strategy
        self.group_size = group_size
        #NOTE: Must keep at least 3 groups if keeping any for propagating
        #      distributed paths.
        self.max_groups = max_groups if max_groups <= 0 else max(3, max_groups)
        self.groups = None  # List[PathGroup]

    @property
    def num_groups(self) -> int:
        ''' Get the number of groups '''
        return len(self.groups) if self.groups is not None else 0

    def add_path(self, path: Path) -> PathGroup:
        '''
        Add the path to an existing group, or create a new group.
        :param path: The path to add
        :return: each closed PathGroup, otherwise None.
        '''
        result = None
        if self.groups is None or len(self.groups) == 0:
            self.groups = [
                PathGroup(
                    accept_strategy = self.accept_strategy,
                    first_path=path,
                )
            ]
        else:
            # Assume always "incremental", such that new paths will belong in
            # the latest group
            latest_group = self.groups[-1]
            if not latest_group.accept(path):
                # Time to add a new group
                self.close_group()
                result = latest_group

                # Start a new group
                self.groups.append(
                    PathGroup(
                        accept_strategy = self.accept_strategy,
                        first_path=path,
                    )
                )

                # Enforce max_group limit by removing groups from the front
                if self.max_groups > 0 and len(self.groups) >= self.max_groups:
                    while len(self.groups) >= self.max_groups:
                        self.groups.pop(0)
        return result

    def close_group(self, idx: int = -1, check_size: bool = True) -> PathGroup:
        '''
        Close the (last) group by
          * Adding distributable lines from the prior group
          * Checking size constraints and dropping if warranted
        :return: The closed PathGroup
        '''
        if self.groups is None or len(self.groups) == 0:
            return None

        if idx == -1:
            idx = len(self.groups) - 1
        latest_group = self.groups[idx]

        # Add distributable lines from the prior group
        if idx > 0:
            latest_group.incorporate_paths(self.groups[idx-1])

        # Check size constraints
        if check_size and self.group_size > 0 and len(self.groups) > 0:
            if latest_group.size < self.group_size:
                # Last group not "filled" ... need to drop
                print('POP!')
                self.groups.pop()

        return latest_group

    def accept_path(self, path: Path):
        '''
        Add the path only if accepted by an existing group.
        :param path: The path to add
        :return: True if accepted
        '''
        for group in self.groups:
            if group.accept(path):
                return True
        return False

    def all_groups_have_size(self, group_size: int) -> bool:
        '''
        :param group_size: The group size to test for.
        :return: True if all groups have the group_size
        '''
        if self.groups is not None:
            for group in self.groups:
                if group.size != group_size:
                    return False
            return True
        return False


class RecordPathBuilder:
    '''
    Class for building record paths from json_data.
    '''
    def __init__(
            self,
            json_data: str,
            output_stream,
            line_builder_fn: Callable[[int, int, str, Any], str],
            timeout: int = TIMEOUT,
    ):
        '''
        :param json_data: The json data (url, file_path, or str)
        :param output_stream: The output stream to write lines to
        :param line_builder_fn: The function for turning record tuples into a line
            string.
        :param timeout: The requests timeout (if json_data is a url)
        '''
        self.jdata = json_data
        self.output_stream = output_stream
        self.builder_fn = line_builder_fn
        self.timeout = timeout
        self.rec_id = 0
        self.inum = 0
        self.sorter = None

    def write_group(self, group: PathGroup):
        for path in group.paths:
            line = self.builder_fn(
                self.rec_id, path.line_num, path.jq_path, path.item
            )
            print(line, file=self.output_stream)

    def visitor(self, item, path):
        jq_path = build_jq_path(path, keep_list_idxs=True)
        group = self.sorter.add_path(Path(jq_path, item, line_num=self.inum))
        if group is not None:
            self.write_group(group)
            self.rec_id += 1
        self.inum += 1

    def stream_record_paths(self):
        self.rec_id = 0
        self.inum = 0
        self.sorter = PathSorter(
            ArrayElementAcceptStrategy(max_array_level=0),
            max_groups=2,
        )

        stream_json_data(self.jdata, self.visitor, timeout=self.timeout)

        last_group = self.sorter.close_group(check_size=False)
        if last_group is not None:
            self.write_group(last_group)
        

def stream_record_paths(
        json_data: str,
        output_stream,
        line_builder_fn: Callable[[int, int, str, Any], str],
        timeout: int = TIMEOUT,
):
    '''
    Write built lines from (rec_id, item_num, jq_path, item) tuples of the
    top-level json "records" to the output stream where:
        * rec_id is a 0-based integer identifying unique records
        * line_num is a 0-based integer identifying the original item number
        * jq_path is the fully-qualified (indexed) path (a record attribute)
        * item is the item value at the path.
    :param json_data: The json data (url, file_path, or str)
    :param output_stream: The output stream to write lines to
    :param line_builder_fn: The function for turning record tuples into a line
        string.
    :param timeout: The requests timeout (if json_data is a url)
    '''
    rpb = RecordPathBuilder(
        json_data, output_stream, line_builder_fn, timeout=timeout
    )
    rpb.stream_record_paths()


def get_records_df(
        json_data: str,
        timeout: int = TIMEOUT
) -> pd.DataFrame:
    '''
    Convenience function to collect the (top-level) record lines from a json
    stream as a DataFrame with columns:
        record_id, line_num, jq_path, item

    WARNING: Don't use this method with very large streams.
    :param json_data: The source json data
    :param timeout: The requests timeout (if json_data is a url)
    :return: The record lines as a dataframe
    '''
    s = io.StringIO()
    stream_record_paths(
        json_data,
        s,
        lambda rid, lid, jqp, val: f'{rid}\t{lid}\t{jqp}\t{val}',
        timeout=timeout,
    )
    s.seek(0)
    df = pd.read_csv(s, sep='\t', names=['rec_id', 'line_num', 'jq_path', 'item'])
    return df



import json
import math
import random
import time
from datetime import datetime, timedelta
from threading import Lock


def wait_for_random_millis(uptomillis):
    waittime = random.uniform(1, uptomillis)
    time.sleep(waittime / 1000)
    return waittime


class StatsAccumulator:
    '''
    A low-memory helper class to collect statistical samples and provide
    summary statistics.
    '''

    def __init__(self, label='', other=None, as_dict=None, values=None):
        '''
        :param label: A label or name for this instance.
        :param other: An other instance for copying into this
        :param as_dict: An "as_dict" form of stats to start with
        :param values: A list of initial values or a single value to add
        '''
        self.label = label
        self._n = 0
        self._min = 0.0
        self._max = 0.0
        self._sum = 0.0
        self._sos = 0.0
        self._modlock = Lock()

        if other is not None:
            if not label == '':
                self._label = other._label
            self._n = other._n
            self._min = other._min
            self._max = other._max
            self._sum = other._sum
            self._sos = other._sos
        if as_dict is not None:
            self.initialize(label=label, as_dict=as_dict)

        if values is not None:
            if isinstance(values, list):
                self.add(*values)
            else:
                self.add(values)

    @property
    def label(self):
        ''' Get the label or name '''
        return self._label

    @label.setter
    def label(self, val):
        ''' Set the label or name '''
        self._label = val

    @property
    def n(self):
        ''' Get the number of values added '''
        return self._n

    @property
    def min(self):
        ''' Get the minimum value added '''
        return self._min

    @property
    def max(self):
        ''' Get the maximum value added '''
        return self._max

    @property
    def sum(self):
        ''' Get the sum of all values added '''
        return self._sum

    @property
    def sum_of_squares(self):
        ''' Get the sum of all squared values '''
        return self._sos

    @property
    def mean(self):
        ''' Get the mean of the values '''
        return 0 if self._n == 0 else self._sum / self._n

    @property
    def std(self):
        ''' Get the standard deviation of the values '''
        return 0 if self._n < 2 else math.sqrt(self.var)

    @property
    def var(self):
        ''' Get the variance of the values '''
        var = 0
        if self._n > 1:
            var = abs(
                (1.0 / (self._n - 1.0)) *
                (self._sos - (1.0 / self._n) * self._sum * self._sum)
            )
        return var

    def clear(self, label=''):
        ''' Clear all values (reset) '''
        self._modlock.acquire()
        try:
            self._label = label
            self._n = 0
            self._min = 0.0
            self._max = 0.0
            self._sum = 0.0
            self._sos = 0.0
        finally:
            self._modlock.release()

    def initialize(self, label='', n=0, min=0, max=0, mean=0, std=0, as_dict=None):
        '''
        Initialize with the given values, preferring existing values from the dictionary.
        '''
        if as_dict is not None:
            if 'label' in as_dict:
                label = as_dict['label']
            if 'n' in as_dict:
                n = as_dict['n']
            if 'min' in as_dict:
                min = as_dict['min']
            if 'max' in as_dict:
                max = as_dict['max']
            if 'mean' in as_dict:
                mean = as_dict['mean']
            if 'std' in as_dict:
                std = as_dict['std']
                

        self._modlock.acquire()
        try:
            self._label = label
            self._n = n
            self._min = min
            self._max = max
            if as_dict is not None and 'sum' in as_dict:
                self._sum = as_dict['sum']
            else:
                self._sum = mean * n
            if as_dict is not None and 'sos' in as_dict:
                self._sos = as_dict['sos']
            else:
                self._sos = 0 if n == 0 else std * std * (n - 1.0) + self._sum * self._sum / n
        finally:
            self._modlock.release()

    def as_dict(self, with_sums=False):
        '''
        Get a dictionary containing a summary of this instance's information.
        '''
        d = {
            'label': self.label,
            'n': self.n,
            'min': self.min,
            'max': self.max,
            'mean': self.mean,
            'std': self.std
        }
        if with_sums:
            d.update({
                'sum': self._sum,
                'sos': self._sos
            })
        return d

    def __str__(self):
        ''' Get the info as a json string '''
        return json.dumps(self.as_dict(), sort_keys=True)

    def add(self, *values):
        ''' Add the value(s) (thread-safe) '''
        self._modlock.acquire()
        try:
            for value in values:
                self._do_add(value)
        finally:
            self._modlock.release()
        return self

    def _do_add(self, value):
        ''' Do the work of adding a value '''
        if self._n == 0:
            self._min = value
            self._max = value
        else:
            if value < self._min:
                self._min = value
            if value > self._max:
                self._max = value

        self._n += 1
        self._sos += (value * value)
        self._sum += value

    @staticmethod
    def combine(label, *stats_accumulators):
        '''
        Create a new statsAccumulator as if it had accumulated all data from
        the given list of stats accumulators.
        '''
        result = StatsAccumulator(label)
        for stats in stats_accumulators:
            result.incorporate(stats)
        return result

    def incorporate(self, other):
        '''
        Incorporate the other stats_accumulator's data into this as if this had
        accumulated the other's along with its own.
        '''
        if other is None:
            return
        
        self._modlock.acquire()
        try:
            if self._n == 0:
                self._min = other.min
                self._max = other.max
            else:
                if other.min < self._min:
                    self._min = other.min
                if other.max > self._max:
                    self._max = other.max
                        
            self._n += other.n
            self._sos += other.sum_of_squares
            self._sum += other.sum
        finally:
            self._modlock.release()


class LinearRegression:
    '''
    A low-memory helper class to collect (x,y) samples and compute the
    linear regression.
    '''

    def __init__(self):
        self._n = 0
        self._x_sum = 0.0
        self._y_sum = 0.0
        self._xy_sum = 0.0
        self._xx_sum = 0.0
        self._yy_sum = 0.0

        self._m = None
        self._b = None

    @property
    def n(self):
        return self._n

    @property
    def x_sum(self):
        return self._x_sum

    @property
    def y_sum(self):
        return self._y_sum

    @property
    def xy_sum(self):
        return self._xy_sum

    @property
    def xx_sum(self):
        return self._xx_sum

    @property
    def yy_sum(self):
        return self._yy_sum

    @property
    def m(self):
        if self._m is None:
            self._compute_regression()
        return self._m

    @property
    def b(self):
        if (self._b is None):
            self._compute_regression()
        return self._b

    def get_y(self, x):
        return self.m * x + self.b

    def add(self, x, y):
        self._m = None
        self._b = None
        self._n += 1
        self._x_sum += x
        self._y_sum += y
        self._xy_sum += (x * y)
        self._xx_sum += (x * x)
        self._yy_sum += (y * y)

    def __str__(self):
        return 'y = %.4f x + %.4f' % (self.m, self.b)

    def _compute_regression(self):
        self._m = 0.0
        self._b = 0.0

        denominator = self._n * self._xx_sum - self._x_sum * self._x_sum
        if (denominator != 0):
            m_numerator = self._n * self._xy_sum - self._x_sum * self._y_sum
            b_numerator = self._y_sum * self._xx_sum - self._x_sum * self._xy_sum

            self._m = m_numerator / denominator
            self._b = b_numerator / denominator


class RollingStats:
    '''
    Implementation for a collection of stats through a rolling window of time.
    '''

    def __init__(self, window_width=300000, segment_width=5000):
        self._modlock = Lock()
        self._window_width = window_width
        self._segment_width = segment_width
        self._window_delta = timedelta(milliseconds=window_width)
        self._cumulative_stats = StatsAccumulator("cumulative")

        self._num_segments = int(round(window_width / segment_width))
        self._segment_stats = [StatsAccumulator("segment-" + str(i)) for i in range(self._num_segments)]
        self._starttime = datetime.now()
        self._reftime = self._starttime
        self._cur_segment = 0

    @property
    def window_width(self):
        return self._window_width

    @property
    def num_segments(self):
        return self._num_segments

    @property
    def cur_segment(self):
        result = -1
        self._modlock.acquire()
        try:
            self._inc_to_cur_segment()
            result = self._cur_segment
        finally:
            self._modlock.release()
        return result

    @property
    def last_segment(self):
        return self._cur_segment

    @property
    def start_time(self):
        return self._starttime

    @property
    def ref_time(self):
        return self._reftime

    @property
    def cumulative_stats(self):
        return self._cumulative_stats

    @property
    def window_stats(self):
        '''
        Get the stats for the current window.
        '''
        self._modlock.acquire()
        try:
            self._inc_to_cur_segment()
            result = StatsAccumulator.combine(self.current_label, *self._segment_stats)
        finally:
            self._modlock.release()
        return result

    @property
    def current_label(self):
        return 'Window-%s-%s' % (str(self._reftime), str(self._window_delta))

    def as_dict(self):
        '''
        Get a dictionary containing a summary of this instance's information.
        '''
        result = {'now': str(datetime.now())}

        cumulative_info = {'since': str(self.start_time)}
        self._add_stats_info(self.cumulative_stats, cumulative_info)
        result['cumulative'] = cumulative_info

        window_info = {'width_millis': self.window_width}
        current_window_stats = self.window_stats
        self._add_stats_info(current_window_stats, window_info)
        result['window'] = window_info

        return result

    def __str__(self):
        return json.dumps(self.as_dict(), sort_keys=True)

    def reset(self):
        self._modlock.acquire()
        try:
            self._starttime = datetime.now()
            self._reftime = self._starttime
            self._cur_segment = 0
            for segment in self._segment_stats:
                segment.clear()
            self._cumulative_stats.clear()
        finally:
            self._modlock.release()

    def add(self, *values):
        '''
        Add the current value(s), returning the segment number to which the
        value as added (useful for testing).
        '''
        self._modlock.acquire()
        try:
            self._inc_to_cur_segment()
            self._segment_stats[self._cur_segment].add(*values)
            result = self._cur_segment
            self._cumulative_stats.add(*values)
        finally:
            self._modlock.release()

        return result

    def has_window_activity(self):
        '''
        Determines whether the current window has activity, returning
            (has_activity, current_window_stats)
        '''
        window_stats = self.window_stats
        return (window_stats.n > 0, window_stats)

    def _inc_to_cur_segment(self):
        result = datetime.now()
        seg_num = int((self._get_millis(result, self._starttime) % self._window_width) / self._segment_width)
        
        diff = self._get_millis(result, self._reftime)

        # if advanced to new segment or wrapped around current
        if seg_num != self._cur_segment or diff > self._segment_width:
            if self._num_segments == 1:
                # special case: wrapped around one and only segment in window
                self._segment_stats[0].clear()
            elif diff > self._window_width:
                # wrapped around the entire window, need to clear all
                for stats in self._segment_stats:
                    stats.clear()
                self._cur_segment = seg_num
            else:
                # walk up to and including new current segment, clearing each
                next_seg_num = (seg_num + 1) % self._num_segments
                i = (self._cur_segment + 1) % self._num_segments
                while i != next_seg_num:
                    self._segment_stats[i].clear()
                    i = (i + 1) % self._num_segments
                self._cur_segment = seg_num

        self._reftime = result
    
    def _get_millis(self, laterdatetime, earlierdatetime):
        return int((laterdatetime - earlierdatetime).total_seconds() * 1000)

    def _add_stats_info(self, stats, info):
        '''
        Add stats summary information to the 'info' dict.
        '''
        if stats.n > 0:
            info['status'] = 'active'
            info['stats'] = stats.as_dict()
            info['millis_per_item'] = self.get_millis_per_item(stats)
            info['items_per_milli'] = self.get_items_per_milli(stats)
        else:
            info['status'] = 'inactive'

    @staticmethod
    def get_millis_per_item(stats):
        '''
        Extract the (average) number of milliseconds per item from the stats.
        '''
        result = None

        if stats is not None:
            if stats.n > 0:
                result = stats.mean

        return result

    @staticmethod
    def get_items_per_milli(stats):
        '''
        Extract the (average) number of items per millisecond from the stats.
        '''
        result = None

        if stats is not None:
            if stats.mean > 0:
                result = 1.0 / stats.mean

        return result


class Monitor:
    '''
    A monitor tracks processing and/or access times and rates for some function.
    '''

    def __init__(
            self,
            description=None,
            access_times=None,
            processing_times=None,
            default_window_width=300000,
            default_segment_width=5000
    ):
        '''
        NOTE: access_times, processing_times should be RollingStats instances.
        '''
        self._modlock = Lock()
        self._alive_since = datetime.now()
        self._description = description
        self._last_start_time = None
        self._access_times = access_times
        self._processing_times = processing_times

        # Defaults for when creating RollingStats from within this Monitor
        self._default_window_width = default_window_width
        self._default_segment_width = default_segment_width


    @property
    def alive_since(self):
        return self._alive_since

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, val):
        self._description = val

    @property
    def last_access_time(self):
        '''
        Get the time at which access was last recorded.
        '''
        return self._last_start_time

    @property
    def access_times(self):
        '''
        Get the access_times (RollingStats).
        '''
        return self._access_times

    @property
    def processing_times(self):
        '''
        Get the processing_times (RollingStats).
        '''
        return self._processing_times

    @property
    def access_cumulative_stats(self):
        return None if self._access_times is None else self._access_times.cumulative_stats

    @property
    def access_window_stats(self):
        return None if self._access_times is None else self._access_times.window_stats

    @property
    def access_window_width(self):
        return None if self._access_times is None else self._access_times.window_width

    @property
    def processing_cumulative_stats(self):
        return None if self._processing_times is None else self._processing_times.cumulative_stats

    @property
    def processing_window_stats(self):
        return None if self._processing_times is None else self._processing_times.window_stats

    @property
    def processing_window_width(self):
        return None if self._processing_times is None else self._processing_times.window_width

    @property
    def default_window_width(self):
        return self._default_window_width

    @default_window_width.setter
    def default_window_width(self, val):
        self._default_window_width = val

    @property
    def default_segment_width(self):
        return self._default_segment_width

    @default_segment_width.setter
    def default_segment_width(self, val):
        self._default_segment_width = val


    def as_dict(self):
        '''
        Get a dictionary containing a summary of this instance's information.
        '''
        result = {}

        result['alive_since'] = str(self.alive_since)
        if self.description is not None:
            result['description'] = self.description
        if self.last_access_time is not None:
            result['last_mark'] = str(self.last_access_time)
        if self.access_times is not None:
            result['access_stats'] = self.access_times.as_dict()
        if self.processing_times is not None:
            result['processing_stats'] = self.processing_times.as_dict()

        return result

    def __str__(self):
        return json.dumps(self.as_dict(), sort_keys=True)

    def get_stats(self, access=False, window=False):
        '''
        Get window/cumulative access/processing stats.
        '''
        result = None

        rolling_stats = self._access_times if access else self._processing_times
        if rolling_stats is not None:
            result = rolling_stats.window_stats if window else rolling_stats.cumulative_stats

        return result

    def mark(self, starttime, endtime=None):
        '''
        Mark another access time, and processing time if endtime is not None.
        '''
        self._modlock.acquire()
        try:
            if self._last_start_time is not None:
                if self._access_times is None:
                    # initialize if needed
                    self._access_times = RollingStats(self.default_window_width, self.default_segment_width)

                self._access_times.add(self._get_millis(starttime, self._last_start_time))

            if endtime is not None:
                if self._processing_times is None:
                    # initialize if needed
                    self._processing_times = RollingStats(self.default_window_width, self.default_segment_width)

                self._processing_times.add(self._get_millis(endtime, starttime))
                    
            self._last_start_time = starttime

        finally:
            self._modlock.release()

    def _get_millis(self, laterdatetime, earlierdatetime):
        return int((laterdatetime - earlierdatetime).total_seconds() * 1000)


class MonitorManager:
    '''
    Class to manage a set of monitors by label, providing rollup views across.
    '''

    def __init__(self, default_window_width=300000, default_segment_width=5000):
        self._monitors = {}
        self._key_manager = KeyManager()

        # Defaults for when creating a Monitor from within this MonitorManager
        self._default_window_width = default_window_width
        self._default_segment_width = default_segment_width


    @property
    def default_window_width(self):
        return self._default_window_width

    @default_window_width.setter
    def default_window_width(self, val):
        self._default_window_width = val

    @property
    def default_segment_width(self):
        return self._default_segment_width

    @default_segment_width.setter
    def default_segment_width(self, val):
        self._default_segment_width = val

    def get_monitors(self):
        return self._monitors

    def get_monitor(self, label, create_if_missing=False, description=None):
        result = None

        if label in self._monitors:
            result = self._monitors[label]

        elif create_if_missing:
            result = Monitor(description=description, default_window_width=self.default_window_width, default_segment_width=self.default_segment_width)
            self._monitors[label] = result

        return result

    def get_or_create_monitor_by_key_type(self, keytype, description):
        key = self._key_manager.get_key(keytype, description)
        return self.get_monitor(key, create_if_missing=True, description=description)

    def set_monitor(self, label, monitor):
        self._monitors[label] = monitor

    def get_stats(self, label=None, access=False, window=False):
        '''
        Get window/cumulative access/processing stats for label or all.
        '''
        result = None

        if label is None:
            result = StatsAccumulator('rollup')

            # Combine access stats across all monitors
            for monitor in self._monitors.values():
                result.incorporate(monitor.get_stats(access, window))

        elif label in self._monitors:
            result = self._monitors[label].get_stats(access, window)

        return result

    def as_dict(self):
        '''
        Get a dictionary containing a summary of this instance's information.
        '''
        result = {}

        # add overall processing/access, cumulative/window stats
        result['overall_stats'] = self.get_overall_stats()

        # add as_dict for each individual monitor
        for key, monitor in self._monitors.items():
            result[key] = monitor.as_dict()

        return result

    def __str__(self):
        return json.dumps(self.as_dict(), sort_keys=True)

    def get_overall_stats(self):
        result = {}

        cumulative_processing = self.get_stats(access=False, window=False)
        if cumulative_processing is not None:
            result['cumulative_processing'] = cumulative_processing.as_dict()

        cumulative_access = self.get_stats(access=True, window=False)
        if cumulative_access is not None:
            result['cumulative_access'] = cumulative_access.as_dict()

        window_processing = self.get_stats(access=False, window=True)
        if window_processing is not None:
            result['window_processing'] = window_processing.as_dict()

        window_access = self.get_stats(access=True, window=True)
        if window_access is not None:
            result['window_access'] = window_access.as_dict()

        return result


class KeyManager:
    '''
    Class to turn descriptions of key types into keys of the form type-N.
    '''

    def __init__(self):
        self._keytype2descriptions = {}

    def get_key(self, keytype, description):

        if keytype in self._keytype2descriptions:
            descriptions = self._keytype2descriptions[keytype]
        else:
            descriptions = []
            self._keytype2descriptions[keytype] = descriptions

        if description in descriptions:
            index = descriptions.index(description)
        else:
            index = len(descriptions)
            descriptions.append(description)

        result = '%s-%d' % (keytype, index)
        return result



import math
import re
from itertools import product
from typing import List, Set


# squash whitespace: to collapse consecutive whitespace to a single space by
#    x.sub(' ', text)
SQUASH_WS_RE = re.compile(r'\s+')


# to identify strings with any symbols by
#    x.search(text)
ALL_SYMBOLS_RE = re.compile(r'[^\w\s]+')


# camelcase LU: to split between consecutive lower and upper chars by
#    x.sub(r'\1 \2', text)
CAMELCASE_LU_RE = re.compile(r'([a-z]+)([A-Z])')


# camelcase UL: to split between consecutive upper and uppler lower chars by
#    x.sub(r'\1 \2', text)
CAMELCASE_UL_RE = re.compile(r'([A-Z]+)([A-Z][a-z])')


# non-embedded symbols: those without a word char on both sides by
#    x.sub('', text)
NON_EMBEDDED_WORD_SYMS_RE = re.compile(r'((?<!\w)[^\w\s]+)|([^\w\s]+(?!\w))')


# embedded symbols: to drop embedded symbols by
#    x.sub('', text)
EMBEDDED_SYMS_RE = re.compile(r'(?<=\w)[^\w\s]+(?=\w)')


# hyphen-slash: to split between an embedded hyphen and/or slash by
#    x.split(text)
HYPHEN_SLASH_RE = re.compile(r'(?<=\w)[\-\/ ](?=\w)')


# hyphen-only: to split between an embedded hyphen by
#    x.split(text)
HYPHEN_ONLY_RE = re.compile(r'(?<=\w)[\- ](?=\w)')


# slash-only: to split between an embedded slash by
#    x.split(text)
SLASH_ONLY_RE = re.compile(r'(?<=\w)\/(?=\w)')


# parenthetical expressions: to drop parenthetical expressions by
#    x.sub('', text)
PARENTHETICAL_RE = re.compile(r'\(.*\)')


# ampersand: to replace an ampersand with " and " by
#    x.sub(' and ', text)
AMPERSAND_RE = re.compile(r'\s*\&\s*')


def expand_camelcase_fn(text: str) -> str:
    '''
    Expand both "lU" and "UUl" camelcasing to "l U" and "U Ul"
    '''
    text = CAMELCASE_LU_RE.sub(r'\1 \2', text)
    return CAMELCASE_UL_RE.sub(r'\1 \2', text)


def drop_non_embedded_symbols_fn(text: str, repl: str = '') -> str:
    '''
    Drop symbols not embedded within word characters
    '''
    return NON_EMBEDDED_WORD_SYMS_RE.sub(repl, text)


def drop_embedded_symbols_fn(text: str, repl: str = '') -> str:
    '''
    Drop symbols embedded within word characters
    '''
    return EMBEDDED_SYMS_RE.sub(repl, text)


def get_hyphen_slash_expansions_fn(
        text: str,
        subs: List[str] = ('-', ' ', ''),
        add_self: bool = True,
        do_split: bool = True,
        min_split_token_len: int = 2,
        hyphen_slash_re=HYPHEN_SLASH_RE,
) -> Set[str]:
    '''
    Given text with words that may or may not appear as hyphenated or with a
    slash, return the set potential variations:
        - the text as-is (add_self)
        - with a hyphen between all words (if '-' in subs)
        - with a space between all words (if ' ' in subs)
        - with all words squashed together (empty string between if '' in subs)
        - with each word separately (do_split as long as min_split_token_len is
              met for all tokens)

    :param text: The hyphen-worthy snippet of text, either already
        hyphenated or with a slash or space delimited.
    :param subs: A string of characters or list of strings to insert between
        tokens
    :param add_self: True to include the text itself in the result
    :param min_split_token_len: If any of the split tokens fail
        to meet the min token length, don't add any of the splits.
    :param hyphen_slash_re: The regex to identify hyphen/slash to expand.

    NOTEs:
      * To add a variation with a slash, add '/' to subs.
      * To not add any variations with symbols, leave them out of subs
        * and don't add self.
    '''
    variations = {text} if add_self else set()
    if subs is not None and len(subs) > 0:
        # create variant with all <s>'s
        for s in subs:
            variations.add(HYPHEN_SLASH_RE.sub(s, text))
    if do_split:
        # add each word separately
        tokens = set(hyphen_slash_re.split(text))
        if not max(map(lambda t: len(t) < min_split_token_len,  tokens)):
            variations.update(tokens)
    return variations


def drop_parentheticals_fn(text: str) -> str:
    '''
    Drop parenthetical expressions from the text.
    '''
    return PARENTHETICAL_RE.sub('', text)
        

def expand_ampersand_fn(text: str) -> str:
    '''
    Replace '&' with ' and '.
    '''
    return AMPERSAND_RE.sub(' and ', text)


def get_lexical_variations(
        text: str,
        include_self: bool = True,
        expand_camelcase: bool = True,
        drop_non_embedded_symbols: bool = True,
        drop_embedded_symbols: bool = True,
        spacify_embedded_symbols: bool = False,
        do_hyphen_expansion: bool = True,
        hyphen_subs: List[str] = (' ', ''),
        do_hyphen_split: bool = True,
        min_hyphen_split_token_len = 2,
        do_slash_expansion: bool = True,
        slash_subs: List[str] = (' ', ' or '),
        do_slash_split: bool = True,
        min_slash_split_token_len: int = 1,
        drop_parentheticals: bool = True,
        expand_ampersands: bool = True,
        add_eng_plurals: bool = True,
) -> Set[str]:
    '''
    Get all variations for the text (including the text itself).
    '''
    variations = {text} if include_self else set()
    if expand_camelcase:
        variations.add(expand_camelcase_fn(text))
    if drop_non_embedded_symbols:
        variations.add(drop_non_embedded_symbols_fn(text))
    if drop_embedded_symbols:
        variations.add(drop_embedded_symbols_fn(text))
    if spacify_embedded_symbols:
        variations.add(drop_embedded_symbols_fn(text, ' '))
    if (
            (
                do_hyphen_expansion and
                hyphen_subs is not None and
                len(hyphen_subs) > 0
            )
            or
            do_hyphen_split
    ):
        variations.update(get_hyphen_slash_expansions_fn(
            text,
            subs=hyphen_subs,
            add_self=False,
            do_split=do_hyphen_split,
            min_split_token_len=min_hyphen_split_token_len,
        ))
    if (
            (
                do_slash_expansion and
                slash_subs is not None and
                len(slash_subs) > 0
            )
            or
            do_slash_split
    ):
        variations.update(get_hyphen_slash_expansions_fn(
            text,
            subs=slash_subs,
            add_self=False,
            do_split=do_slash_split,
            min_split_token_len=min_slash_split_token_len,
        ))
    if drop_parentheticals:
        variations.add(drop_parentheticals_fn(text))
    if expand_ampersands:
        variations.add(expand_ampersand_fn(text))
    if add_eng_plurals:
        # TODO: Use a better pluralizer
        plurals = {
            f'{v}s' for v in variations
        }
        variations.update(plurals)
    return variations


def int_to_en(num: int) -> str:
    d = { 0 : 'zero', 1 : 'one', 2 : 'two', 3 : 'three', 4 : 'four', 5 : 'five',
          6 : 'six', 7 : 'seven', 8 : 'eight', 9 : 'nine', 10 : 'ten',
          11 : 'eleven', 12 : 'twelve', 13 : 'thirteen', 14 : 'fourteen',
          15 : 'fifteen', 16 : 'sixteen', 17 : 'seventeen', 18 : 'eighteen',
          19 : 'nineteen', 20 : 'twenty',
          30 : 'thirty', 40 : 'forty', 50 : 'fifty', 60 : 'sixty',
          70 : 'seventy', 80 : 'eighty', 90 : 'ninety' }
    k = 1000
    m = k * 1000
    b = m * 1000
    t = b * 1000

    if not isinstance(num, int):
        return num

    if num < 0:
        return 'negative ' + int_to_en(abs(num))

    if (num < 20):
        return d[num]

    if (num < 100):
        if num % 10 == 0: return d[num]
        else: return d[num // 10 * 10] + ' ' + d[num % 10]

    if (num < k):
        if num % 100 == 0: return d[num // 100] + ' hundred'
        else: return d[num // 100] + ' hundred and ' + int_to_en(num % 100)

    if (num < m):
        if num % k == 0: return int_to_en(num // k) + ' thousand'
        else: return int_to_en(num // k) + ' thousand ' + int_to_en(num % k)

    if (num < b):
        if (num % m) == 0: return int_to_en(num // m) + ' million'
        else: return int_to_en(num // m) + ' million ' + int_to_en(num % m)

    if (num < t):
        if (num % b) == 0: return int_to_en(num // b) + ' billion'
        else: return int_to_en(num // b) + ' billion ' + int_to_en(num % b)

    if (num % t == 0): return int_to_en(num // t) + ' trillion'
    else: return int_to_en(num // t) + ' trillion ' + int_to_en(num % t)

    # num is too large
    return str(num)


def zero_pad_variations(
        val: int,
        min_zpad_len: int,
        max_zpad_len: int,
) -> Set[str]:
    '''
    Get (only) zero-padded variations of the given value from min (inclusive)
    to max (exclusive) zero-pad lengths.

    Examples:
      * zero_pad_variations(9, 2, 4) == {'09', '009'}
      * zero_pad_variations(90, 2, 4) == {'090'}
      * zero_pad_variations(90, 2, 3) == {}
      * zero_pad_variations(3, 0, 5) == {'03', '003', '0003'}

    :param min_zpad_len: The minimum zero-padded string length (inclusive)
    :param max_zpad_len: The maximum zero-padded string length (exclusive)
    :return: The set of all requested zero-padded number strings
    '''
    return {
        f'{val:0{zpad}d}'
        for zpad in range(
                max(
                    min_zpad_len,
                    math.ceil(math.log10(val)) + 1 if val > 0 else 1
                ),
                max_zpad_len
        )
    }


def month_day_variations_fn(
        month_or_day: int,
        do_int_to_en: bool = False,
) -> Set[str]:
    '''
    Get the variations for a month or day number, including the number
    itself as a string, a 2-digit zero-padded form of the number, and
    (optionally) english word for the number.

    :param month_or_day: The month or day for which to get variations
    :param do_int_to_en: Optionally include the english word for the number
    :return: The set of variations for the value
    '''
    result = zero_pad_variations(month_or_day, 2, 3)
    result.add(str(month_or_day))
    if do_int_to_en:
        result.add(int_to_en(month_or_day))
    return result


def year_variations_fn(
        year: int,
        min_year: int = 0,
        max_year: int = 9999,
        do_int_to_en_below_100: bool = False,
        numeric_only: bool = False,
) -> Set[str]:
    '''
        * "1999"
        * Convert a year to english text:
          * Long text: one thousand, nine hundred and ninety nine
          * Short text: nineteen [hundred and] ninety nine
    '''
    variations = {str(year)}

    if year < min_year or year > max_year:
        return variations

    # one thousand, nine hundred and ninety nine
    if not numeric_only and (do_int_to_en_below_100 or year >= 100):
        variations.add(int_to_en(year))

    # nineteen ninety five
    century = year // 100
    remainder = year % 100
    remainder_text = int_to_en(remainder)

    variations.update(zero_pad_variations(remainder, 2, 3))

    if century > 0:
        remainder_texts = list()
        if remainder > 0:
            if remainder < 10:
                if not numeric_only:
                    remainder_texts.append(f' oh {remainder_text}')
                remainder_texts.append(f' 0{remainder}')
            else:
                if not numeric_only:
                    remainder_texts.append(f' {remainder_text}')
                remainder_texts.append(f' {remainder}')
            if not numeric_only:
                remainder_texts.append(f' and {remainder_text}')

        century_text = int_to_en(century)
        scales = ['', century_text]
        if century % 10 == 0:
            mil_text = int_to_en(century // 10)
            scales.append(f'{mil_text} thousand')
        else:
            scales.append(f'{century_text} hundred')

        def clean_up(s):
            s = s.strip()
            if s.startswith('and '):
                s = s[4:]
            return s

        variations.update({
            clean_up(''.join(v))
            for v in product(scales, remainder_texts)
        })

    return variations


def replace_smart_quotes_fn(text: str) -> str:
    '''
    Replace "smart" quotes with their ascii version.
    '''
    return text.replace(
        '\u201c', '"'   # left double quote U+201C
    ).replace(
        '\u201d', '"'   # right double quote U+201D
    ).replace(
        '\u2018', "'"   # left single quote U+2018
    ).replace(
        '\u2019', "'"   # right single quote U+2019
    )


def basic_normalization_fn(
        text: str,
        lowercase: bool = True,
        expand_camelcase: bool = True,
        simplify_quote_chars: bool = True,
        drop_non_embedded_symbols: bool = False,
        spacify_embedded_symbols: bool = False,
        drop_embedded_symbols: bool = False,
        squash_whitespace: bool = False,
        do_all: bool = False,
) -> str:
    '''
    Basic normalization functions include:
        * lowercasing [default]
        * expanding camelcase [default]
        * replacing "smart" quotes and apostrophes with ascii verisons [default]
        * dropping non_embedded symbols [optional]
        * replacing embedded symbols with a space [takes precedence over dropping unless do_all]
        * or dropping embedded symbols [optional]
        * collapsing multiple spaces and stripping spaces from ends [optional]
    '''
    # NOTE: do this before changing case
    if expand_camelcase or do_all:
        text = expand_camelcase_fn(text)

    if lowercase or do_all:
        text = text.lower()
    if (drop_non_embedded_symbols and drop_embedded_symbols) or do_all:
        text = re.sub(r'[^\w\s]+', '', text)
    elif drop_non_embedded_symbols:
        text = drop_non_embedded_symbols_fn(text)
    elif spacify_embedded_symbols:
        text = drop_embedded_symbols_fn(text, ' ')
    elif drop_embedded_symbols:
        text = drop_embedded_symbols_fn(text)

    # NOTE: do this after dropping (only some) symbols
    if simplify_quote_chars and (
            not drop_non_embedded_symbols or
            not drop_embedded_symbols
    ):
        # NOTE: It only makes sense to do this if we're keeping symbols
        text = replace_smart_quotes_fn(text)

    # NOTE: do this last
    if squash_whitespace or do_all:
        text = re.sub(r'\s+', ' ', text).strip()
    return text



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dataknobs', 'dataknobs.structures', 'dataknobs.utils', 'dataknobs.xization']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.11.1,<5.0.0',
 'json-stream>=2.1.1,<3.0.0',
 'lxml>=4.9.2,<5.0.0',
 'more-itertools>=9.0.0,<10.0.0',
 'nltk>=3.7,<4.0',
 'numpy>=1.23.3,<2.0.0',
 'pandas>=2.1.1,<3.0.0',
 'psycopg2-binary>=2.9.3,<3.0.0',
 'python-dotenv>=0.21.0,<0.22.0',
 'requests>=2.28.1,<3.0.0',
 'scikit-learn>=1.2.1,<2.0.0']

setup_kwargs = {
    'name': 'dataknobs',
    'version': '0.0.8',
    'description': 'Useful implementations of data structures and design patterns for AI knowledge bases.',
    'long_description': 'DataKnobs\n=============================\n\n## Description\n\nUseful implementations of data structures and design patterns for knowledge bases and AI, or the knobs and levers for fine-tuning and leveraging your data.\n\nThis repo also serves as a template or sandbox for development, experimentation, and testing of general data structures, algorithms, and utilities for DS, AI, ML, and NLP.\n\nProvides connectors for other popular text and data processing packages like:\n  * numpy and pandas\n  * nltk\n  * wordnet\n  * postgres\n  * elasticsearch\n\n## General project information\n\nThe purpose of this project is:\n\n  * To provide dependable implementations of useful data structures.\n  * To show examples of design patterns and ways to apply AI concepts.\n  * To prototype tools for delivering a robust DS/AI/ML/NLP utilities library package.\n  * To facilitate interactive development, demonstration, visualization, and testing of the library components via jupter notebooks and/or scripts.\n\n## Installation and Usage\n\n```bash/python\n% pip install dataknobs\n% python\n>>> import dataknobs as dk\n>>> ...\n```\n\n\n## Development\n\n### Development machine prerequisites\n\nThe following minimum configuration should exist for development:\n\n  * tox\n  * pyenv\n     * pyenv install 3.9\n  * poetry\n\nWith optional:\n\n  * docker\n  * bash\n\nBy convention, a data directory can be leveraged for development that is mounted as a shared volumne in Docker as /data. This has the default of $HOME/data, but can be overridden with the DATADIR environment variable.\n\n\n### Development quickstart guide\n\n  * In a terminal, clone the repo and cd into the project directory.\n\n#### Testing\n\n  * Tests and Lint: "tox"\n  * Just unit tests: "tox -e tests"\n  * Just lint: "tox -e lint"\n\n#### Using docker\n\n  * Development:\n```\n% tox -e dev\n# poetry shell\n# python\n```\n\n  * Notebook:\n    * execute "tox -e nb"\n      * copy/paste url into browser\n\n#### Using virtual environments\n\n  * Development:\n    * Manual: source ".project_vars", poetry install, poetry shell\n    * Automated: execute "bin/start_dev.sh"  (requires "/bin/bash" on your machine)\n\n  * Notebook:\n    * execute "bin/start_notebook.sh"\n      * copy/paste url into browser\n',
    'author': 'Spence Koehler',
    'author_email': 'KoehlerSB747@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.13',
}


setup(**setup_kwargs)



import gzip
import os
from typing import Set


def filepath_generator(
        rootpath: str,
        descend: bool = True,
        seen: Set[str] = None,
        files_only: bool = True,
):  # yields -> str
    '''
    Generate all filepaths under the root path.

    :param rootpath: The root path under which to find files.
    :param descend: True to descend into subdirectories
    :param seen: Set of filepaths and/or directories to ignore
    :param files_only: True to generate only paths to files; False to
        include paths to directories.
    :yield: Each file path
    '''
    if seen is None:
        seen = set()
    seen.add(rootpath)
    for root, dirs, files in os.walk(rootpath, topdown=True):
        if not descend and root != rootpath and root in seen:
            break
        for name in files:
            fpath = os.path.join(root, name)
            if fpath not in seen:
                seen.add(fpath)
                yield fpath
        if not descend or not files_only:
            for name in dirs:
                next_root = os.path.join(root, name)
                if next_root not in seen:
                    seen.add(next_root)
                    yield next_root


def fileline_generator(filename: str, rootdir: str = None):
    '''
    Generate lines from the file.
    :param filename: The name of the file.
    :param rootdir: (optional) The directory of the file.
    :yield: Each stripped file line
    '''
    if filename.endswith('.gz'):
        open_fn = gzip.open
        mode = 'rt'
    else:
        open_fn = open
        mode = 'r'
    if rootdir is not None:
        filename = os.path.join(rootdir, filename)
    with open_fn(filename, mode=mode, encoding='utf-8') as f:
        for line in f:
            yield line.strip()


def write_lines(outfile, lines, rootdir=None):
    '''
    Write the lines to the file.
    :param outfile: The name of the file.
    :param rootdir: (optional) The directory of the file.
    '''
    if rootdir is not None:
        outfile = os.path.join(rootdir, outfile)
    if outfile.endswith('.gz'):
        open_fn = gzip.open
        mode = 'wt'
    else:
        open_fn = open
        mode = 'w'
    with open_fn(outfile, mode=mode, encoding='utf-8') as f:
        for line in sorted(lines):
            print(line, file=f)


def is_gzip_file(filepath: str) -> bool:
    '''
    Determine whether the file at filepath is gzipped.
    :param filepath: The path to the file
    :return: True if the file is gzipped
    '''
    is_gzip = False
    if os.path.exists(filepath):
        with open(filepath, 'rb') as f:
            b = f.read(3)
            is_gzip = (b == b'\x1f\x8b\x08')
    return is_gzip



import json
import requests
import socket
from typing import Any, Callable, Dict, Tuple


DEFAULT_TIMEOUT = 5
HEADERS = {"Content-Type": "application/json"}
DBG_HEADERS = {"Content-Type": "application/json", "error_trace": "true"}


def get_current_ip() -> str:
    '''
    Get the running machine's IPv4 address.
    :return: The IP address
    '''
    return socket.gethostbyname(socket.gethostname())


def json_api_response_handler(
        resp: requests.models.Response
) -> Tuple[int, Dict[str, Any]]:
    result = None
    if resp.text:
        result = json.loads(resp.text)
    return (resp, result)


def plain_api_response_handler(
        resp: requests.models.Response
) -> Tuple[int, Dict[str, Any]]:
    return (resp, resp.text)


default_api_response_handler = json_api_response_handler


def get_request(
        api_request: str,
        params: Dict[str, Any] = None,
        headers: Dict[str, Any] = None,
        timeout: int = DEFAULT_TIMEOUT,
        api_response_handler: Callable[
            requests.models.Response, Tuple[int, Dict[str, Any]]
        ] = default_api_response_handler,
        requests=requests,  # pylint: disable-msg=W0621
) -> Tuple[int, Dict[str, Any]]:
    '''
    Submit the api get request and collect the response as a Dict.
    :param api_request: The api request
    :param params: The request parameters
    :param headers: The request headers (defaults to HEADERS)
    :param timeout: The request timeout (in seconds)
    :param api_response_handler: A handler function for api responses
    :param requests: Arg for alternate or "mock" requests package override
    :return: The response code and the json result as a dict (or None)
    '''
    if headers is None:
        headers = HEADERS
    return api_response_handler(
        requests.get(api_request, headers=headers, params=params, timeout=timeout)
    )


def post_request(
        api_request: str,
        payload: Dict[str, Any],
        params: Dict[str, Any] = None,
        headers: Dict[str, Any] = None,
        timeout: int = DEFAULT_TIMEOUT,
        api_response_handler: Callable[
            requests.models.Response, Tuple[int, Dict[str, Any]]
        ] = default_api_response_handler,
        requests=requests,  # pylint: disable-msg=W0621
) -> Tuple[int, Dict[str, Any]]:
    '''
    Submit the api post request and collect the response as a Dict.
    :param api_request: The api request
    :param params: The request parameters
    :param headers: The request headers (defaults to HEADERS)
    :param timeout: The request timeout (in seconds)
    :param api_response_handler: A handler function for api responses
    :param requests: Arg for alternate or "mock" requests package override
    :return: The response code and the json result as a dict (or None)
    '''
    if headers is None:
        headers = HEADERS
    return api_response_handler(
        requests.post(
            api_request, data=payload, headers=headers, params=params,
            timeout=timeout,
        )
    )


def post_files_request(
        api_request: str,
        files: Dict[str, Any],
        headers: Dict[str, Any] = None,
        timeout: int = DEFAULT_TIMEOUT,
        api_response_handler: Callable[
            requests.models.Response, Tuple[int, Dict[str, Any]]
        ] = default_api_response_handler,
        requests=requests,  # pylint: disable-msg=W0621
) -> Tuple[int, Dict[str, Any]]:
    '''
    Post data from one or more files.
    :param api_request: The api request
    :param files: A dict of {<file_id>: <file_data>}} entries for each file,
        where files can be of the form of the following examples:
            1. Just file data:
               - {'myfile': open('report.xls', 'rb')}  # NOTE: open in binary mode!
            2. Including filename, content_type, and headers
               - {'myfile': (
                                'report.xls',
                                open('report.xls', 'rb'),
                                'application/vnd.ms-excel',
                                {'Expires': '0'}
                            )}
    :param headers: The request headers
    :param timeout: The request timeout (in seconds)
    :param api_response_handler: A handler function for api responses
    :param requests: Arg for alternate or "mock" requests package override
    :return: The response code and the json result as a dict (or None)
    '''
    return api_response_handler(
        requests.post(api_request, files=files, headers=headers, timeout=timeout)
    )


def put_request(
        api_request: str,
        payload: Dict[str, Any],
        params: Dict[str, Any] = None,
        headers: Dict[str, Any] = None,
        timeout: int = DEFAULT_TIMEOUT,
        api_response_handler: Callable[
            requests.models.Response, Tuple[int, Dict[str, Any]]
        ] = default_api_response_handler,
        requests=requests,  # pylint: disable-msg=W0621
) -> Tuple[int, Dict[str, Any]]:
    '''
    Submit the api put request and collect the response as a Dict.
    :param api_request: The api request
    :param params: The request parameters
    :param headers: The request headers (defaults to HEADERS)
    :param timeout: The request timeout (in seconds)
    :param api_response_handler: A handler function for api responses
    :param requests: Arg for alternate or "mock" requests package override
    :return: The response code and the json result as a dict (or None)
    '''
    if headers is None:
        headers = HEADERS
    return api_response_handler(
        requests.put(
            api_request, data=payload, headers=headers, params=params,
            timeout=timeout,
        )
    )


def delete_request(
        api_request: str,
        params: Dict[str, Any] = None,
        headers: Dict[str, Any] = None,
        timeout: int = DEFAULT_TIMEOUT,
        api_response_handler: Callable[
            requests.models.Response, Tuple[int, Dict[str, Any]]
        ] = default_api_response_handler,
        requests=requests,  # pylint: disable-msg=W0621
) -> Tuple[int, Dict[str, Any]]:
    '''
    Submit the api delete request and collect the response as a Dict.
    :param api_request: The api request
    :param params: The request parameters
    :param headers: The request headers (defaults to HEADERS)
    :param timeout: The request timeout (in seconds)
    :param api_response_handler: A handler function for api responses
    :param requests: Arg for alternate or "mock" requests package override
    :return: The response code and the json result as a dict (or None)
    '''
    if headers is None:
        headers = HEADERS
    return api_response_handler(
        requests.delete(api_request, headers=headers, params=params, timeout=timeout)
    )


class ServerResponse:
    '''
    Class to encapsulate request response data from the elasticsearch server.
    '''

    def __init__(self, resp, result):
        self.resp = resp
        self.result = result
        self._extra = None

    def __repr__(self):
        rv = ''
        if self.result:
            rv = f'({self.status}):\n{json.dumps(self.result, indent=2)}'
        else:
            rv = str(self.status)
        return rv
        

    @property
    def succeeded(self):
        return self.resp.status_code == 200 if self.resp is not None else False

    @property
    def status(self):
        return self.resp.status_code if self.resp is not None else None

    @property
    def extra(self):
        if self._extra is None:
            self._extra = dict()
        return self._extra

    def has_extra(self):
        return self._extra is not None and len(self._extra) > 0

    def add_extra(self, key, value):
        self.extra[key] = value


class RequestHelper:
    '''
    Class to simplify sending api request commands to a server.
    '''

    def __init__(
            self, server_ip, server_port,
            api_response_handler=json_api_response_handler,
            headers=None,
            timeout=DEFAULT_TIMEOUT,
            mock_requests=None,
    ):
        self.ip = server_ip
        self.port = server_port
        self.response_handler = api_response_handler
        self.headers = headers
        self.timeout = timeout
        self.requests = mock_requests if mock_requests else requests

    def build_url(self, path):
        return f'http://{self.ip}:{self.port}/{path}'

    def request(
            self,
            rtype, path, payload=None,
            params=None, files=None,
            response_handler=None,
            headers=None,
            timeout=0,
            verbose=True,
    ):
        '''
        :param rtype: The request type, or command. One of:
            ['get', 'post', 'post-files', 'put', 'delete']
        :param path: The api path portion of the request
        :param payload: The request payload
        :param params: The request params
        :param files: The request files
        :param response_handler: Response handler to override instance value
        :param headers: Headers to override instance value
        :param timeout: The request timeout override (in seconds) if not 0
        :param verbose: True to print server response info
        :return: A ServerResponse instance with the results
        '''
        rtype = rtype.lower()
        if timeout == 0:
            timeout = self.timeout
        if headers is None:
            headers = self.headers
        if response_handler is None:
            response_handler = self.response_handler
        url = self.build_url(path)
        resp, result = None, None
        if rtype == 'get':
            resp, result = get_request(
                url, params=params, headers=headers, timeout=timeout,
                api_response_handler=response_handler if response_handler is not None else self.response_handler,
                requests=self.requests,
            )
        elif rtype == 'post':
            resp, result = post_request(
                url, payload,
                params=params, headers=headers, timeout=timeout,
                api_response_handler=response_handler,
                requests=self.requests,
            )
        elif rtype == 'post-files':
            resp, result = post_files_request(
                url, files=files, headers=headers, timeout=timeout,
                api_response_handler=response_handler,
                requests=self.requests,
            )
        elif rtype == 'put':
            resp, result = put_request(
                url, payload,
                params=params, headers=headers, timeout=timeout,
                api_response_handler=response_handler,
                requests=self.requests,
            )
        elif rtype == 'delete':
            resp, result = delete_request(
                url, params=params, headers=headers, timeout=timeout,
                api_response_handler=response_handler,
                requests=self.requests,
            )
    
        rv = ServerResponse(resp, result)
    
        if verbose:
            print(rv)
    
        return rv


class MockResponse:
    '''
    A mock response object
    '''
    def __init__(self, status_code, result):
        self.status_code = status_code
        self.result = result
        self.text = result
        if not isinstance(result, str):
            self.text = json.dumps(result)

    def to_server_response(self):
        ''' Convenience method for creating a ServerResponse '''
        return ServerResponse(self, self.result)


class MockRequests:
    def __init__(self):
        self.responses = dict()
        self.r404 = MockResponse(404, '"Not found"')

    def add(
            self,
            response, api, api_request,
            data=None, files=None,
            headers=None, params=None,
            timeout=DEFAULT_TIMEOUT,
    ):
        key = self._make_key(
            api, api_request, data, files, headers, params, timeout,
        )
        self.responses[key] = response


    def _make_key(
            self,
            api, api_request,
            data, files,
            headers, params,
            timeout,
    ):
        return json.dumps({
            'api': api,
            'req': api_request,
            'data': data,
            'files': files,
            'headers': headers,
            'params': params,
            'timeout': timeout,
        })

    def get(self, api_request, headers=None, params=None, timeout=None):
        key = self._make_key(
            'get', api_request, None, None, headers, params, timeout
        )
        return self.responses.get(key, self.r404)

    def post(self, api_request, data=None, files=None, headers=None, params=None, timeout=None):
        key = self._make_key('post', api_request, data, files, headers, params, timeout)
        return self.responses.get(key, self.r404)

    def put(self, api_request, data=None, headers=None, params=None, timeout=None):
        key = self._make_key('put', api_request, data, None, headers, params, timeout)
        return self.responses.get(key, self.r404)

    def delete(self, api_request, headers=None, params=None, timeout=None):
        key = self._make_key('delete', api_request, None, None, headers, params, timeout)
        return self.responses.get(key, self.r404)
